""" Special Numerical Operations. """

from __future__ import division

import numpy as np


#
# Module constants
#

EPS = np.finfo(float).eps
TINY = np.finfo(float).tiny
SMALL = 1e-100
LOGTINY = np.log(TINY)


#
# Special functions
#

def logsumexp(X, axis=0):
    """
    Log-sum-exp trick for matrix X for summation along a specified axis.

    This performs the following operation in a stable fashion,

    .. math::

        \log \sum^K_{k=1} \exp\{x_k\}

    Parameters
    ----------
        X: ndarray
            2D array of shape (N, D) to apply the log-sum-exp trick.
        axis: int, optional
            Axis to apply the summation along (works the same as axis in
            numpy.sum).

    Returns
    -------
        lseX: ndarray
            results of applying the log-sum-exp trick, this will be shape (D,)
            if :code:`axis=0` or shape (N,) if :code:`axis=1`.
    """

    mx = X.max(axis=axis)
    if (X.ndim > 1):
        mx = np.atleast_2d(mx).T if axis == 1 else np.atleast_2d(mx)

    return np.log(np.exp(X - mx).sum(axis=axis)) + np.ravel(mx)


def softmax(X, axis=0):
    """
    Pass X through a softmax function in a numerically stable way using the
    log-sum-exp trick.

    This transformation is:

    .. math::

        \\frac{\exp\{X_k\}}{\sum^K_{j=1} \exp\{X_j\}}

    and is appliedx to each row/column, `k`, of X.

    Parameters
    ----------
        X: ndarray
            2D array of shape (N, D) to apply the log-sum-exp trick.
        axis: int, optional
            Axis to apply the summation along (works the same as axis in
            numpy.sum).

    Returns
    -------
        smX: ndarray
            results of applying the log-sum-exp trick, this will be shape
            (N, D), and each row will sum to 1 if :code:`axis=1` or each column
            will sum to 1 if :code:`axis=0`.
    """

    if axis == 1:
        return np.exp(X - logsumexp(X, axis=1)[:, np.newaxis])
    elif axis == 0:
        return np.exp(X - logsumexp(X, axis=0))
    else:
        raise ValueError("This only works on 2D arrays for now.")


def softplus(X):
    """ Pass X through a soft-plus function, , in a numerically
        stable way (using the log-sum-exp trick).

        The softplus transformation is:

        .. math::
            \log(1 + \exp\{X\})

        Parameters
        ----------
            X: ndarray
                shape (N,) array or shape (N, D) array of data.

        Returns
        -------
            spX: ndarray
                array of same shape of X with the result of softmax(X).
    """

    if np.isscalar(X):
        return logsumexp(np.vstack((np.zeros(1), [X])).T, axis=1)[0]

    N = X.shape[0]

    if X.ndim == 1:
        return logsumexp(np.vstack((np.zeros(N), X)).T, axis=1)
    elif X.ndim == 2:
        sftX = np.empty(X.shape, dtype=float)
        for d in range(X.shape[1]):
            sftX[:, d] = logsumexp(np.vstack((np.zeros(N), X[:, d])).T, axis=1)
        return sftX
    else:
        raise ValueError("This only works on up to 2D arrays.")


#
# Numerically "safe" functions
#

def safelog(x, min_x=TINY):

    cx = x.copy()
    cx[cx < min_x] = min_x
    return np.log(cx)


def safesoftplus(x, min_x=SMALL):

    g = softplus(x)
    g[g < min_x] = min_x
    return g



"""
Reusable utility functions
"""

from .decorators import (vectorize_args, unvectorize_args, vectorize_result)
from .base import (flatten, unflatten, couple, decouple, nwise, map_indices,
                   atleast_list, atleast_tuple, issequence, map_recursive)
from .rand import endless_permutations

__all__ = [
    'vectorize_args',
    'unvectorize_args',
    'vectorize_result',
    'flatten',
    'unflatten',
    'couple',
    'decouple',
    'nwise',
    'map_indices',
    'atleast_list',
    'atleast_tuple',
    'issequence',
    'map_recursive',
    'endless_permutations'
]



"""
The :mod:`revrand.optimize` module provides a standardized interface to
``scipy.optimize``, and also supports custom optimization methods.
"""

from .sgd import sgd, AdaDelta, AdaGrad, Momentum, Adam, SGDUpdater
from .decorators import (structured_minimizer,
                         structured_sgd,
                         logtrick_minimizer,
                         logtrick_sgd)

__all__ = [
    'sgd',
    'AdaDelta',
    'AdaGrad',
    'Momentum',
    'Adam',
    'SGDUpdater',
    'structured_minimizer',
    'structured_sgd',
    'logtrick_minimizer',
    'logtrick_sgd'
]



"""Assortment of handy functions."""

import inspect
import numpy as np

from six.moves import map, range, zip
from functools import partial
from itertools import tee
from inspect import isgenerator


class Bunch(dict):
    """
    Container object for datasets.

    Dictionary-like object that exposes its keys as attributes.

    Examples
    --------
    >>> b = Bunch(foo=42, bar=10)
    >>> b == {'foo': 42, 'bar': 10}
    True
    >>> b.foo
    42
    >>> b.bar
    10
    >>> b['foo']
    42
    >>> b.baz = 61
    >>> b.baz
    61
    >>> b['baz']
    61
    """

    def __init__(self, **kwargs):
        dict.__init__(self, kwargs)
        self.__dict__ = self


def issequence(obj):
    """
    Test if an object is an iterable generator, list or tuple.

    Parameters
    ----------
    obj : object
        object to test

    Returns
    -------
    bool :
        True if :code:`obj` is a tuple, list or generator only.

    Examples
    --------
    >>> issequence([1, 2])
    True
    >>> issequence((1,))
    True
    >>> issequence((i for i in range(8)))
    True
    >>> issequence(np.array([1, 2, 3]))
    False
    """
    return inspect.isgenerator(obj) or isinstance(obj, (list, tuple))


def atleast_list(a):
    """
    Promote an object to a list if not a list or generator.

    Parameters
    ----------
    a: object
        any object you want to at least be a list with one element

    Returns
    -------
    list or generator:
        untounched if :code:`a` was a generator or list, otherwise :code:`[a]`.

    Examples
    --------
    >>> a = 1.
    >>> atleast_list(a)
    [1.0]
    >>> a = [1.]
    >>> atleast_list(a)
    [1.0]
    """
    return a if isinstance(a, list) or isgenerator(a) else [a]


def atleast_tuple(a):
    """
    Promote an object to a tuple if not a tuple or generator.

    Parameters
    ----------
    a: object
        any object you want to at least be a tuple with one element

    Returns
    -------
    tuple or generator:
        untounched if :code:`a` was a generator or tuple, otherwise
        :code:`(a,)`.

    Examples
    --------
    >>> a = 1.
    >>> atleast_tuple(a)
    (1.0,)
    >>> a = (1.,)
    >>> atleast_tuple(a)
    (1.0,)
    """
    return a if isinstance(a, tuple) or isgenerator(a) else (a,)


def couple(f, g):
    r"""
    Compose a function thate returns two arguments.

    Given a pair of functions that take the same arguments, return a
    single function that returns a pair consisting of the return values
    of each function.

    Notes
    -----
    Equivalent to::

        lambda f, g: lambda *args, **kwargs: (f(*args, **kwargs),
                                              g(*args, **kwargs))

    Examples
    --------
    >>> f = lambda x: 2*x**3
    >>> df = lambda x: 6*x**2
    >>> f_new = couple(f, df)
    >>> f_new(5)
    (250, 150)

    """
    def coupled(*args, **kwargs):
        return f(*args, **kwargs), g(*args, **kwargs)
    return coupled


def decouple(fn):
    """
    Inverse operation of couple.

    Create two functions of one argument and one return from a function that
    takes two arguments and has two returns

    Examples
    --------
    >>> h = lambda x: (2*x**3, 6*x**2)
    >>> f, g = decouple(h)

    >>> f(5)
    250

    >>> g(5)
    150
    """
    def fst(*args, **kwargs):
        return fn(*args, **kwargs)[0]

    def snd(*args, **kwargs):
        return fn(*args, **kwargs)[1]

    return fst, snd


def nwise(iterable, n):
    r"""
    Sliding window iterator.

    Iterator that acts like a sliding window of size `n`; slides over
    some iterable `n` items at a time. If iterable has `m` elements,
    this function will return an iterator over `m-n+1` tuples.

    Parameters
    ----------
    iterable : iterable
        An iterable object.

    n : int
        Window size.

    Returns
    -------
    iterator of tuples.
        Iterator of size `n` tuples

    Notes
    -----
    First `n` iterators are created::

        iters = tee(iterable, n)

    Next, iterator `i` is advanced `i` times::

        for i, it in enumerate(iters):
            for _ in range(i):
                next(it, None)

    Finally, the iterators are zipped back up again::

        return zip(*iters)

    Examples
    --------
    >>> a = [2, 5, 7, 4, 2, 8, 6]

    >>> list(nwise(a, n=3))
    [(2, 5, 7), (5, 7, 4), (7, 4, 2), (4, 2, 8), (2, 8, 6)]

    >>> pairwise = partial(nwise, n=2)
    >>> list(pairwise(a))
    [(2, 5), (5, 7), (7, 4), (4, 2), (2, 8), (8, 6)]

    >>> list(nwise(a, n=1))
    [(2,), (5,), (7,), (4,), (2,), (8,), (6,)]

    >>> list(nwise(a, n=7))
    [(2, 5, 7, 4, 2, 8, 6)]

    .. todo::

       These should probably raise `ValueError`...

    >>> list(nwise(a, 8))
    []

    >>> list(nwise(a, 9))
    []

    A sliding window of size `n` over a list of `m` elements
    gives `m-n+1` windows

    >>> len(a) - len(list(nwise(a, 2))) == 1
    True

    >>> len(a) - len(list(nwise(a, 3))) == 2
    True

    >>> len(a) - len(list(nwise(a, 7))) == 6
    True
    """
    iters = tee(iterable, n)
    for i, it in enumerate(iters):
        for _ in range(i):
            next(it, None)
    return zip(*iters)


def scalar_reshape(a, newshape, order='C'):
    """
    Reshape, but also return scalars or empty lists.

    Identical to `numpy.reshape` except in the case where `newshape` is
    the empty tuple, in which case we return a scalar instead of a
    0-dimensional array.

    Examples
    --------
    >>> a = np.arange(6)
    >>> np.array_equal(np.reshape(a, (3, 2)), scalar_reshape(a, (3, 2)))
    True

    >>> scalar_reshape(np.array([3.14]), newshape=())
    3.14

    >>> scalar_reshape(np.array([2.71]), newshape=(1,))
    array([ 2.71])

    >>> scalar_reshape(np.array([]), newshape=(0,))
    []
    """
    if newshape == ():
        return np.asscalar(a)

    if newshape == (0,):
        return []

    return np.reshape(a, newshape, order)


def flatten(arys, returns_shapes=True, hstack=np.hstack, ravel=np.ravel,
            shape=np.shape):
    """
    Flatten a potentially recursive list of multidimensional objects.

    .. note::

       Not to be confused with `np.ndarray.flatten()` (a more befitting
       might be `chain` or `stack` or maybe something else entirely
       since this function is more than either `concatenate` or
       `np.flatten` itself. Rather, it is the composition of the former
       with the latter.

    Parameters
    ----------
    arys : list of objects
        One or more input arrays of possibly heterogenous shapes and
        sizes.
    returns_shapes : bool, optional
        Default is `True`. If `True`, the tuple `(flattened, shapes)` is
        returned, otherwise only `flattened` is returned.
    hstack : callable, optional
        a function that implements horizontal stacking
    ravel : callable, optional
        a function that flattens the object
    shape : callable, optional
        a function that returns the shape of the object

    Returns
    -------
    flattened,[shapes] : {1dobject, list of tuples}
        Return the flat (1d) object resulting from the concatenation of
        flattened multidimensional objects. When `returns_shapes` is `True`,
        return a list of tuples containing also the shapes of each array as the
        second element.

    See Also
    --------
    revrand.utils.unflatten : its inverse

    Examples
    --------
    >>> a = 9
    >>> b = np.array([4, 7, 4, 5, 2])
    >>> c = np.array([[7, 3, 1],
    ...               [2, 6, 6]])
    >>> d = np.array([[[6, 5, 5],
    ...                [1, 6, 9]],
    ...               [[3, 9, 1],
    ...                [9, 4, 1]]])

    >>> flatten([a, b, c, d]) # doctest: +NORMALIZE_WHITESPACE
    (array([9, 4, 7, 4, 5, 2, 7, 3, 1, 2, 6, 6, 6, 5, 5, 1, 6, 9, 3, 9,
            1, 9, 4, 1]), [(), (5,), (2, 3), (2, 2, 3)])

    Note that scalars and 0-dimensional arrays are treated differently
    from 1-dimensional singleton arrays.

    >>> flatten([3.14, np.array(2.71), np.array([1.61])])
    ... # doctest: +NORMALIZE_WHITESPACE
    (array([ 3.14,  2.71,  1.61]), [(), (), (1,)])

    >>> flatten([a, b, c, d], returns_shapes=False)
    ... # doctest: +NORMALIZE_WHITESPACE
    array([9, 4, 7, 4, 5, 2, 7, 3, 1, 2, 6, 6, 6, 5, 5, 1, 6, 9, 3, 9,
           1, 9, 4, 1])

    >>> w, x, y, z = unflatten(*flatten([a, b, c, d]))

    >>> w == a
    True

    >>> np.array_equal(x, b)
    True

    >>> np.array_equal(y, c)
    True

    >>> np.array_equal(z, d)
    True

    >>> flatten([3.14, [np.array(2.71), np.array([1.61])]])
    ... # doctest: +NORMALIZE_WHITESPACE
    (array([ 3.14,  2.71,  1.61]), [(), [(), (1,)]])

    """
    if issequence(arys) and len(arys) > 0:

        flat = partial(flatten,
                       returns_shapes=True,
                       hstack=hstack,
                       ravel=ravel,
                       shape=shape
                       )

        flat_arys, shapes = zip(*map(flat, arys))
        flat_ary = hstack(flat_arys)
        shapes = list(shapes)

    else:

        flat_ary = ravel(arys)
        shapes = shape(arys)

    return (flat_ary, shapes) if returns_shapes else flat_ary


def unflatten(ary, shapes, reshape=scalar_reshape):
    r"""
    Inverse opertation of flatten.

    Given a flat (1d) array, and a list of shapes (represented as tuples),
    return a list of ndarrays with the specified shapes.

    Parameters
    ----------
    ary : a 1d array
        A flat (1d) array.

    shapes : list of tuples
        A list of ndarray shapes (tuple of array dimensions)

    Returns
    -------
    list of ndarrays
        A list of ndarrays with the specified shapes.

    See Also
    --------
    revrand.utils.flatten : its inverse

    Notes
    -----
    Equivalent to::

        lambda ary, shapes, order='C': \
            map(partial(custom_reshape, order=order),
                np.hsplit(ary, np.cumsum(map(partial(np.prod, dtype=int),
                                             shapes))), shapes)

    Examples
    --------

    >>> a = np.array([7, 4, 5, 8, 9, 1, 4, 2, 5, 3, 4, 3])

    >>> list(unflatten(a, [(1,), (1,), (4,), (2, 3)]))
    ... # doctest: +NORMALIZE_WHITESPACE
    [array([7]), array([4]), array([5, 8, 9, 1]), array([[4, 2, 5],
        [3, 4, 3]])]

    >>> list(unflatten(a, [(), (1,), (4,), (2, 3)]))
    ... # doctest: +NORMALIZE_WHITESPACE
    [7, array([4]), array([5, 8, 9, 1]), array([[4, 2, 5], [3, 4, 3]])]

    >>> list(unflatten(a, [(), (1,), (3,), (2, 3)]))
    ... # doctest: +NORMALIZE_WHITESPACE
    [7, array([4]), array([5, 8, 9]), array([[1, 4, 2], [5, 3, 4]])]

    >>> list(unflatten(a, [(), (1,), (5,), (2, 3)]))
    ... # doctest: +NORMALIZE_WHITESPACE
    Traceback (most recent call last):
        ...
    ValueError: total size of new array must be unchanged

    >>> flatten(list(unflatten(a, [(), (1,), (4,), (2, 3)])))
    ... # doctest: +NORMALIZE_WHITESPACE
    (array([7, 4, 5, 8, 9, 1, 4, 2, 5, 3, 4, 3]),
        [(), (1,), (4,), (2, 3)])

    >>> list(unflatten(a, [[(1,), (1,)], (4,), (2, 3)]))
    ... # doctest: +NORMALIZE_WHITESPACE
    [[array([7]), array([4])], array([5, 8, 9, 1]), array([[4, 2, 5],
        [3, 4, 3]])]

    >>> flatten(list(unflatten(a, [(), (1,), [(4,), (2, 3)]])))
    ... # doctest: +NORMALIZE_WHITESPACE
    (array([7, 4, 5, 8, 9, 1, 4, 2, 5, 3, 4, 3]),
        [(), (1,), [(4,), (2, 3)]])
    """
    if isinstance(shapes, list):
        sizes = list(map(sumprod, shapes))
        ends = np.cumsum(sizes)
        begs = np.concatenate(([0], ends[:-1]))
        struct_arys = [unflatten(ary[b:e], s, reshape=reshape)
                       for b, e, s in zip(begs, ends, shapes)]
        return struct_arys
    else:
        struct_ary = reshape(ary, shapes)
        return struct_ary


def sumprod(seq):
    """
    Product of tuple, or sum of products of lists of tuples.

    Parameters
    ----------
    seq : tuple or list

    Returns
    -------
    int :
        the product of input tuples, or the sum of products of lists of tuples,
        recursively.

    Examples
    --------
    >>> tup = (1, 2, 3)
    >>> sumprod(tup)
    6

    >>> lis = [(1, 2, 3), (2, 2)]
    >>> sumprod(lis)
    10

    >>> lis = [(1, 2, 3), [(2, 1), (3,)]]
    >>> sumprod(lis)
    11
    """
    if isinstance(seq, tuple):
        # important to make sure dtype is int
        # since prod on empty tuple is a float (1.0)
        return np.prod(seq, dtype=int)
    else:
        return np.sum((sumprod(s) for s in seq), dtype=int)


def map_recursive(fn, iterable, output_type=None):
    """
    Apply a function of a potentially nested list of lists.

    Parameters
    ----------
    fn : callable
        The function to apply to each element (and sub elements) in iterable
    iterable : iterable
        An iterable, sequence, sequence of sequences etc. :code:`fn` will be
        applied to each element in each list.
    output_type : callable, optional
        if None, a map with sub-maps in the same structure as :code:`iterable`
        will be returned, otherwise the callable will be applied to each
        sequence (i.e. :code:`list` will return lists of lists etc).

    Returns
    -------
    map or iterable type :
        if :code:`output_type` is None, a map with sub-maps in the same
        structure as :code:`iterable` will be returned, otherwise the callable
        will be applied to each sequence (i.e. :code:`list` will return lists
        of lists etc).

    Examples
    --------
    >>> seq = [1, 2, [3, 4, [5, 6]], 7]
    >>> map_recursive(lambda x: x > 4, seq, output_type=list)
    [False, False, [False, False, [True, True]], True]

    >>> map_recursive(lambda x: 2 * x, seq, output_type=tuple)
    (2, 4, (6, 8, (10, 12)), 14)
    """
    def applyormap(it):
        if issequence(it):
            return map_recursive(fn, it, output_type)
        else:
            return fn(it)

    applied = map(applyormap, iterable)
    return output_type(applied) if output_type else applied


def map_indices(fn, iterable, indices):
    r"""
    Map a function across indices of an iterable.

    Notes
    -----
    Roughly equivalent to, though more efficient than::

        lambda fn, iterable, *indices: (fn(arg) if i in indices else arg
                                        for i, arg in enumerate(iterable))

    Examples
    --------

    >>> a = [4, 6, 7, 1, 6, 8, 2]

    >>> from operator import mul
    >>> list(map_indices(partial(mul, 3), a, [0, 3, 5]))
    [12, 6, 7, 3, 6, 24, 2]

    >>> b = [9., np.array([5., 6., 2.]),
    ...      np.array([[5., 6., 2.], [2., 3., 9.]])]

    >>> list(map_indices(np.log, b, [0, 2])) # doctest: +NORMALIZE_WHITESPACE
    [2.1972245773362196,
     array([ 5.,  6.,  2.]),
     array([[ 1.60943791,  1.79175947,  0.69314718],
            [ 0.69314718,  1.09861229,  2.19722458]])]

    .. todo::

       Floating point precision

    >>> list(map_indices(np.exp, list(map_indices(np.log, b, [0, 2])), [0, 2]))
    ... # doctest: +NORMALIZE_WHITESPACE +SKIP
    [9.,
     array([5., 6., 2.]),
     array([[ 5.,  6.,  2.],
            [ 2.,  3.,  9.]])]
    """
    index_set = set(indices)
    for i, arg in enumerate(iterable):
        if i in index_set:
            yield fn(arg)
        else:
            yield arg



"""
Dataset loading utilities

Portions of this module derived from
http://scikit-learn.org/stable/modules/classes.html#module-sklearn.datasets
"""

import numpy as np
import requests
import tarfile
import os

from six.moves import urllib
from scipy.io import loadmat
from scipy.spatial.distance import cdist
from unipath import Path
from io import BytesIO
from sklearn.utils import check_random_state

from .base import Bunch


def make_regression(func, n_samples=100, n_features=1, bias=0.0, noise=0.0,
                    random_state=None):
    """
    Make dataset for a regression problem.

    Examples
    --------
    >>> f = lambda x: 0.5*x + np.sin(2*x)
    >>> X, y = make_regression(f, bias=.5, noise=1., random_state=1)
    >>> X.shape
    (100, 1)
    >>> y.shape
    (100,)
    >>> X[:5].round(2)
    array([[ 1.62],
           [-0.61],
           [-0.53],
           [-1.07],
           [ 0.87]])
    >>> y[:5].round(2)
    array([ 0.76,  0.48, -0.23, -0.28,  0.83])
    """
    generator = check_random_state(random_state)

    X = generator.randn(n_samples, n_features)
    # unpack the columns of X
    y = func(*X.T) + bias

    if noise > 0.0:
        y += generator.normal(scale=noise, size=y.shape)

    return X, y


def make_polynomial(degree=3, n_samples=100, bias=0.0, noise=0.0,
                    return_coefs=False, random_state=None):
    """
    Generate a noisy polynomial for a regression problem

    Examples
    --------
    >>> X, y, coefs = make_polynomial(degree=3, n_samples=200, noise=.5,
    ...                               return_coefs=True, random_state=1)
    """
    generator = check_random_state(random_state)

    # TODO: Add arguments to support other priors
    coefs = generator.randn(degree + 1)
    pows = np.arange(degree + 1)
    poly = np.vectorize(lambda x: np.sum(coefs * x ** pows))
    X, y = make_regression(poly, n_samples=n_samples, bias=bias, noise=noise,
                           random_state=random_state)
    if return_coefs:
        return X, y, coefs

    return X, y


def get_data_home(data_home=None):
    """
    Return the path of the revrand data dir.

    This folder is used by some large dataset loaders to avoid
    downloading the data several times.

    By default the data dir is set to a folder named 'revrand_data'
    in the user home folder.

    Alternatively, it can be set by the 'REVRAND_DATA' environment
    variable or programmatically by giving an explicit folder path. The
    '~' symbol is expanded to the user home folder.

    If the folder does not already exist, it is automatically created.
    """
    data_home_default = Path(__file__).ancestor(3).child('demos',
                                                         '_revrand_data')

    if data_home is None:
        data_home = os.environ.get('REVRAND_DATA', data_home_default)

    if not os.path.exists(data_home):
        os.makedirs(data_home)

    return data_home


def fetch_gpml_sarcos_data(transpose_data=True, data_home=None):
    """
    Fetch the SARCOS dataset from the internet and parse appropriately into
    python arrays

    >>> gpml_sarcos = fetch_gpml_sarcos_data()

    >>> gpml_sarcos.train.data.shape
    (44484, 21)

    >>> gpml_sarcos.train.targets.shape
    (44484,)

    >>> gpml_sarcos.train.targets.round(2) # doctest: +ELLIPSIS
    array([ 50.29,  44.1 ,  37.35, ...,  22.7 ,  17.13,   6.52])

    >>> gpml_sarcos.test.data.shape
    (4449, 21)

    >>> gpml_sarcos.test.targets.shape
    (4449,)
    """
    train_src_url = "http://www.gaussianprocess.org/gpml/data/sarcos_inv.mat"
    test_src_url = ("http://www.gaussianprocess.org/gpml/data/sarcos_inv_test"
                    ".mat")

    data_home = get_data_home(data_home=data_home)

    train_filename = os.path.join(data_home, 'sarcos_inv.mat')
    test_filename = os.path.join(data_home, 'sarcos_inv_test.mat')

    if not os.path.exists(train_filename):
        urllib.request.urlretrieve(train_src_url, train_filename)

    if not os.path.exists(test_filename):
        urllib.request.urlretrieve(test_src_url, test_filename)

    train_data = loadmat(train_filename).get('sarcos_inv')
    test_data = loadmat(test_filename).get('sarcos_inv_test')

    train_bunch = Bunch(data=train_data[:, :21],
                        targets=train_data[:, 21])

    test_bunch = Bunch(data=test_data[:, :21],
                       targets=test_data[:, 21])

    return Bunch(train=train_bunch, test=test_bunch)


def fetch_gpml_usps_resampled_data(transpose_data=True, data_home=None):
    """
    Fetch the USPS handwritten digits dataset from the internet and parse
    appropriately into python arrays

    >>> usps_resampled = fetch_gpml_usps_resampled_data()

    >>> usps_resampled.train.targets.shape
    (4649,)

    >>> usps_resampled.train.targets # doctest: +ELLIPSIS
    array([6, 0, 1, ..., 9, 2, 7])

    >>> usps_resampled.train.data.shape
    (4649, 256)

    >>> np.all(-1 <= usps_resampled.train.data)
    True

    >>> np.all(usps_resampled.train.data < 1)
    True

    >>> usps_resampled.test.targets.shape
    (4649,)

    >>> usps_resampled.test.data.shape
    (4649, 256)

    >>> usps_resampled = fetch_gpml_usps_resampled_data(transpose_data=False)
    >>> usps_resampled.train.data.shape
    (256, 4649)
    """
    data_home = get_data_home(data_home=data_home)
    data_filename = os.path.join(data_home,
                                 'usps_resampled/usps_resampled.mat')

    if not os.path.exists(data_filename):

        r = requests.get('http://www.gaussianprocess.org/gpml/data/'
                         'usps_resampled.tar.bz2')

        with tarfile.open(fileobj=BytesIO(r.content)) as tar_infile:
            tar_infile.extract('usps_resampled/usps_resampled.mat',
                               path=data_home)

    matlab_dict = loadmat(data_filename)

    train_data = matlab_dict['train_patterns']
    test_data = matlab_dict['test_patterns']

    if transpose_data:
        train_data = train_data.T
        test_data = test_data.T

    train_targets = matlab_dict['train_labels'].T
    train_targets = np.argwhere(train_targets == 1)[:, 1]

    test_targets = matlab_dict['test_labels'].T
    test_targets = np.argwhere(test_targets == 1)[:, 1]

    train_bunch = Bunch(data=train_data,
                        targets=train_targets)

    test_bunch = Bunch(data=test_data,
                       targets=test_targets)

    return Bunch(train=train_bunch, test=test_bunch)


def gen_gausprocess_se(ntrain, ntest, noise=1., lenscale=1., scale=1.,
                       xmin=-10, xmax=10):
    """
    Generate a random (noisy) draw from a Gaussian Process with a RBF kernel.
    """

    # Xtrain = np.linspace(xmin, xmax, ntrain)[:, np.newaxis]
    Xtrain = np.random.rand(ntrain)[:, np.newaxis] * (xmin - xmax) - xmin
    Xtest = np.linspace(xmin, xmax, ntest)[:, np.newaxis]
    Xcat = np.vstack((Xtrain, Xtest))

    K = scale * np.exp(-cdist(Xcat, Xcat, metric='sqeuclidean') /
                       (2 * lenscale**2))
    U, S, V = np.linalg.svd(K)
    L = U.dot(np.diag(np.sqrt(S))).dot(V)
    f = np.random.randn(ntrain + ntest).dot(L)

    ytrain = f[0:ntrain] + np.random.randn(ntrain) * noise
    ftest = f[ntrain:]

    return Xtrain, ytrain, Xtest, ftest



"""
Various linear algebra utilities.

Notes on the fast Hadamard transform (Alistair Reid, NICTA 2015):

The Hadamard transform is a recursive application of sums and differences
on a vector of length 2^n. The log(n) recursive application allow computation
in n log(n) operations instead of the naive n^2 per vector that would result
from computing and then multiplying a Hadamard matrix.

Note: the Walsh ordering is naturally computed by the recursive algorithm. The
sequence ordering requires a re-ordering afterwards incurring a similar
overhead to the original computation.

.. code:: python

    # Larger tests against calling Julia: [length 4*512 vectors:]
    M = np.argsort(np.sin(np.arange(64))).astype(float)
        Julia: 570us (sequence ordered)
        Al's single optimised: 560us (sequence ordered), 30us (natural)
        Al's basic vectorised: 1370us (sequence ordered)
"""

from __future__ import division

import numpy as np
from scipy.linalg import cholesky, cho_solve, svd, LinAlgError


# Numerical constants / thresholds
CHOLTHRESH = 1e-5


def cho_log_det(L):
    """
    Compute the log of the determinant of :math:`A`, given its (upper or lower)
    Cholesky factorization :math:`LL^T`.

    Parameters
    ----------
    L: ndarray
        an upper or lower Cholesky factor

    Examples
    --------
    >>> A = np.array([[ 2, -1,  0],
    ...               [-1,  2, -1],
    ...               [ 0, -1,  2]])

    >>> Lt = cholesky(A)
    >>> np.isclose(cho_log_det(Lt), np.log(np.linalg.det(A)))
    True

    >>> L = cholesky(A, lower=True)
    >>> np.isclose(cho_log_det(L), np.log(np.linalg.det(A)))
    True
    """
    return 2 * np.sum(np.log(L.diagonal()))


def svd_log_det(s):
    """
    Compute the log of the determinant of :math:`A`, given its singular values
    from an SVD factorisation (i.e. :code:`s` from :code:`U, s, Ut = svd(A)`).

    Parameters
    ----------
    s: ndarray
        the singular values from an SVD decomposition

    Examples
    --------
    >>> A = np.array([[ 2, -1,  0],
    ...               [-1,  2, -1],
    ...               [ 0, -1,  2]])

    >>> _, s, _ = np.linalg.svd(A)
    >>> np.isclose(svd_log_det(s), np.log(np.linalg.det(A)))
    True
    """
    return np.sum(np.log(s))


def solve_posdef(A, b):
    """
    Solve the system :math:`A X = b` for :math:`X` where :math:`A` is a
    positive semi-definite matrix.

    This first tries cholesky, and if numerically unstable with solve using a
    truncated SVD (see solve_posdef_svd).

    The log-determinant of :math:`A` is also returned since it requires minimal
    overhead.

    Parameters
    ----------
    A: ndarray
        A positive semi-definite matrix.
    b: ndarray
        An array or matrix

    Returns
    -------
    X: ndarray
        The result of :math:`X = A^-1 b`
    logdet: float
        The log-determinant of :math:`A`
    """
    # Try cholesky for speed
    try:
        lower = False
        L = cholesky(A, lower=lower)
        if any(L.diagonal() < CHOLTHRESH):
            raise LinAlgError("Unstable cholesky factor detected")
        X = cho_solve((L, lower), b)
        logdet = cho_log_det(L)

    # Failed cholesky, use svd to do the inverse
    except LinAlgError:

        U, s, V = svd(A)
        X = svd_solve(U, s, V, b)
        logdet = svd_log_det(s)

    return X, logdet


def svd_solve(U, s, V, b, s_tol=1e-15):
    """
    Solve the system :math:`A X = b` for :math:`X`.

    Here :math:`A` is a positive semi-definite matrix using the singular value
    decomposition. This truncates the SVD so only dimensions corresponding to
    non-negative and sufficiently large singular values are used.

    Parameters
    ----------
    U: ndarray
        The :code:`U` factor of :code:`U, s, V = svd(A)` positive
        semi-definite matrix.
    s: ndarray
        The :code:`s` factor of :code:`U, s, V = svd(A)` positive
        semi-definite matrix.
    V: ndarray
        The :code:`V` factor of :code:`U, s, V = svd(A)` positive
        semi-definite matrix.
    b: ndarray
        An array or matrix
    s_tol: float
        Cutoff for small singular values. Singular values smaller than
        :code:`s_tol` are clamped to :code:`s_tol`.

    Returns
    -------
    X: ndarray
        The result of :math:`X = A^-1 b`
    okind: ndarray
        The indices of :code:`s` that are kept in the factorisation
    """
    # Test shapes for efficient computations
    n = U.shape[0]
    assert(b.shape[0] == n)
    m = b.shape[1] if np.ndim(b) > 1 else 1

    # Auto clamp SVD based on threshold
    sclamp = np.maximum(s, s_tol)

    # Inversion factors
    ss = 1. / np.sqrt(sclamp)
    U2 = U * ss[np.newaxis, :]
    V2 = ss[:, np.newaxis] * V

    if m < n:
        # Few queries
        X = U2.dot(V2.dot(b))  # O(n^2 (2m))
    else:
        X = U2.dot(V2).dot(b)  # O(n^2 (m + n))

    return X


def hadamard(Y, ordering=True):
    """
    *Very fast* Hadamard transform for single vector at a time.

    Parameters
    ----------
    Y: ndarray
        the n*2^k data to be 1d hadamard transformed
    ordering: bool, optional
        reorder from Walsh to sequence

    Returns
    -------
    H: ndarray
        hadamard transformed data.

    Examples
    --------
    from https://en.wikipedia.org/wiki/Hadamard_transform with normalisation

    >>> y = np.array([[1, 0, 1, 0, 0, 1, 1, 0]])
    >>> hadamard(y, ordering=False)
    array([[ 0.5 ,  0.25,  0.  , -0.25,  0.  ,  0.25,  0.  ,  0.25]])
    >>> hadamard(y, ordering=True)
    array([[ 0.5 ,  0.  ,  0.  ,  0.  , -0.25,  0.25,  0.25,  0.25]])
    """
    # dot is a sum product over the last axis of a and the second-to-last of b.
    # Transpose - can specify axes, default transposes a[0] and a[1] hmm
    n_vectors, n_Y = Y.shape
    matching = (n_vectors, 2, int(n_Y / 2))
    H = np.array([[1, 1], [1, -1]]) / 2.  # Julia uses 2 and not sqrt(2)?
    steps = int(np.log(n_Y) / np.log(2))
    assert(2**steps == n_Y)  # required
    for _ in range(steps):
        Y = np.transpose(Y.reshape(matching), (0, 2, 1)).dot(H)
    Y = Y.reshape((n_vectors, n_Y))
    if ordering:
        Y = Y[:, _sequency(n_Y)]
    return Y


def _sequency(length):
    # http://fourier.eng.hmc.edu/e161/lectures/wht/node3.html
    # Although this incorrectly explains grey codes...
    s = np.arange(length).astype(int)
    s = (s >> 1) ^ s  # Grey code ...
    # Reverse bits
    order = np.zeros(s.shape).astype(int)
    n = int(1)
    m = length // 2
    while n < length:
        order |= m * (n & s) // n
        n <<= 1
        m >>= 1
    return order



"""Stochastic Gradient Descent."""

from itertools import chain

import numpy as np
from scipy.optimize import OptimizeResult
# from sklearn.utils import check_random_state

from ..utils import issequence, endless_permutations


#
# SGD updater classes for different learning rate methods
#

class SGDUpdater:
    """
    Base class for SGD learning rate algorithms.

    Parameters
    ----------
    eta : float, optional
        The learning rate applied to every gradient step.
    """

    def __init__(self, eta=0.1):
        """See class docstring."""
        self.eta = eta

    def __call__(self, x, grad):
        """
        Get a new parameter value.

        Parameters
        ----------
        x : ndarray
            input parameters to optimise
        grad : ndarray
            gradient of x

        Returns
        -------
        x_new : ndarray
            the new value for x
        """
        x_new = x - self.eta * grad
        return x_new

    def reset(self):
        """Reset the state of this updater for a new optimisation problem."""
        pass

    def __repr__(self):
        """Representation."""
        return "{}(eta={})".format(self.__class__.__name__, self.eta)


class AdaDelta(SGDUpdater):
    """
    AdaDelta Algorithm.

    Parameters
    ----------
    rho : float, optional
        smoothing/decay rate parameter, must be [0, 1].
    epsilon : float, optional
        "jitter" term to ensure continued learning (should be small).
    """

    def __init__(self, rho=0.1, epsilon=1e-5):
        """See class docstring."""
        if rho < 0 or rho > 1:
            raise ValueError("Decay rate 'rho' must be between 0 and 1!")

        if epsilon <= 0:
            raise ValueError("Constant 'epsilon' must be > 0!")

        self.rho = rho
        self.epsilon = epsilon
        self.Eg2 = 0
        self.Edx2 = 0

    def __call__(self, x, grad):
        """
        Get a new parameter value from AdaDelta.

        Parameters
        ----------
        x : ndarray
            input parameters to optimise
        grad : ndarray
            gradient of x

        Returns
        -------
        x_new : ndarray
            the new value for x
        """
        self.Eg2 = self.rho * self.Eg2 + (1 - self.rho) * grad**2
        dx = - grad * np.sqrt(self.Edx2 + self.epsilon) \
            / np.sqrt(self.Eg2 + self.epsilon)
        self.Edx2 = self.rho * self.Edx2 + (1 - self.rho) * dx**2

        x_new = x + dx
        return x_new

    def reset(self):
        """Reset the state of this updater for a new optimisation problem."""
        self.__init__(self.rho, self.epsilon)

    def __repr__(self):
        """Representation."""
        return "{}(rho={}, epsilon={})".format(
            self.__class__.__name__,
            self.rho,
            self.epsilon
        )


class AdaGrad(SGDUpdater):
    """
    AdaGrad Algorithm.

    Parameters
    ----------
    eta : float, optional
        smoothing/decay rate parameter, must be [0, 1].
    epsilon : float, optional
        small constant term to prevent divide-by-zeros
    """

    def __init__(self, eta=1, epsilon=1e-6):
        """See class docstring."""
        if eta <= 0:
            raise ValueError("Learning rate 'eta' must be > 0!")

        if epsilon <= 0:
            raise ValueError("Constant 'epsilon' must be > 0!")

        self.eta = eta
        self.epsilon = epsilon
        self.g2_hist = 0

    def __call__(self, x, grad):
        """
        Get a new parameter value from AdaGrad.

        Parameters
        ----------
        x : ndarray
            input parameters to optimise
        grad : ndarray
            gradient of x

        Returns
        -------
        x_new : ndarray
            the new value for x
        """
        self.g2_hist += grad**2
        x_new = x - self.eta * grad / (self.epsilon + np.sqrt(self.g2_hist))
        return x_new

    def reset(self):
        """Reset the state of this updater for a new optimisation problem."""
        self.__init__(self.eta, self.epsilon)

    def __repr__(self):
        """Representation."""
        return "{}(eta={}, epsilon={})".format(
            self.__class__.__name__,
            self.eta,
            self.epsilon
        )


class Momentum(SGDUpdater):
    """
    Momentum Algorithm.

    Parameters
    ----------
    rho : float, optional
        smoothing/decay rate parameter, must be [0, 1].
    eta : float, optional
        weight to give to the momentum term
    """

    def __init__(self, rho=0.5, eta=0.01):
        """See class docstring."""
        if eta <= 0:
            raise ValueError("Learning rate 'eta' must be > 0!")

        if rho < 0 or rho > 1:
            raise ValueError("Decay rate 'rho' must be between 0 and 1!")

        self.eta = eta
        self.rho = rho
        self.dx = 0

    def __call__(self, x, grad):
        """
        Get a new parameter value from the Momentum algorithm.

        Parameters
        ----------
        x : ndarray
            input parameters to optimise
        grad : ndarray
            gradient of x

        Returns
        -------
        x_new : ndarray
            the new value for x
        """
        self.dx = self.rho * self.dx - self.eta * grad

        x_new = x + self.dx
        return x_new

    def reset(self):
        """Reset the state of this updater for a new optimisation problem."""
        self.__init__(self.rho, self.eta)

    def __repr__(self):
        """Representation."""
        return "{}(rho={}, eta={})".format(
            self.__class__.__name__,
            self.rho,
            self.eta
        )


class Adam(SGDUpdater):
    """
    Adam Algorithm.

    Parameters
    ----------
    alpha : float, optional
        stepsize to give the update.
    beta1 : float, optional
        smoothing/decay rate parameter for the gradient, must be [0, 1].
    beta2 : float, optional
        smoothing/decay rate parameter for the squared gradient, must be
        [0, 1].
    epsilon : float, optional
        "jitter" term to ensure continued learning (should be small).
    """

    def __init__(self, alpha=0.01, beta1=0.9, beta2=0.99, epsilon=1e-8):
        """See class docstring."""
        self.alpha = alpha
        self.beta1 = beta1
        self.beta2 = beta2
        self.epsilon = epsilon
        self.t = 0
        self.m = None
        self.v = None

    def __call__(self, x, grad):
        """
        Get a new parameter value from the Adam algorithm.

        Parameters
        ----------
        x : ndarray
            input parameters to optimise
        grad : ndarray
            gradient of x

        Returns
        -------
        x_new : ndarray
            the new value for x
        """
        self.t += 1

        if self.m is None:
            self.m = np.zeros_like(x)
            self.v = np.zeros_like(x)

        self.m = self.beta1 * self.m + (1 - self.beta1) * grad
        self.v = self.beta2 * self.v + (1 - self.beta2) * grad**2
        mbar = self.m / (1 - self.beta1**self.t)
        vbar = self.v / (1 - self.beta2**self.t)

        x_new = x - self.alpha * mbar / (np.sqrt(vbar) + self.epsilon)
        return x_new

    def reset(self):
        """Reset the state of this updater for a new optimisation problem."""
        self.__init__(self.alpha, self.beta1, self.beta2, self.epsilon)

    def __repr__(self):
        """Representation."""
        return "{}(alpha={}, beta1={}, beta2={}, epsilon={})".format(
            self.__class__.__name__,
            self.alpha,
            self.beta1,
            self.beta2,
            self.epsilon
        )


#
# SGD minimizer
#

def sgd(fun, x0, data, args=(), bounds=None, batch_size=10, maxiter=5000,
        updater=None, eval_obj=False, random_state=None):
    """
    Stochastic Gradient Descent.

    Parameters
    ----------
    fun : callable
        the function to *minimize*, this must have the signature ``[obj,]``
        grad = fun(x, data, ...)`, where the ``eval_obj`` argument tells
        ``sgd`` if an objective function value is going to be returned by
        ``fun``.
    x0 : ndarray
        a sequence/1D array of initial values for the parameters to learn.
    data : ndarray
        a numpy array or sequence of data to input into ``fun``. This will
        be split along the first axis (axis=0), and then input into
        ``fun``.
    args : sequence, optional
        an optional sequence of arguments to give to fun.
    bounds : sequence, optional
        Bounds for variables, (min, max) pairs for each element in x, defining
        the bounds on that parameter.  Use None for one of min or max when
        there is no bound in that direction.
    batch_size : int, optional
        The number of observations in an SGD batch.
    maxiter : int, optional
        Number of mini-batch iterations before optimization terminates.
    updater : SGDUpdater, optional
        The type of gradient update to use, by default this is Adam.
    eval_obj : bool, optional
        This indicates whether or not ``fun`` also evaluates and returns
        the objective function value. If this is true, ``fun`` must return
        ``(obj, grad)`` and then a list of objective function values is
        also returned.
    random_state : int or RandomState, optional
        random seed

    Returns
    -------
    res : OptimizeResult
        x : narray
            the final result
        norms : list
            the list of gradient norms
        message : str
            the convergence condition ('maxiter reached' or error)
        objs : list
            the list of objective function evaluations if ``eval_obj``
            is True.
        fun : float
            the final objective function evaluation if ``eval_obj`` is
            True.
    """
    if updater is None:
        updater = Adam()

    # Make sure we aren't using a recycled updater
    updater.reset()

    N = _len_data(data)
    x = np.array(x0, copy=True, dtype=float)
    D = x.shape[0]

    # Make sure we have a valid batch size
    batch_size = min(batch_size, N)

    # Process bounds
    if bounds is not None:
        if len(bounds) != D:
            raise ValueError("The dimension of the bounds does not match x0!")

        lower, upper = zip(*map(normalize_bound, bounds))
        lower = np.array(lower)
        upper = np.array(upper)

    # Learning Records
    obj = None
    objs = []
    norms = []

    for batch in gen_batch(data, batch_size, maxiter, random_state):

        if not eval_obj:
            grad = fun(x, *chain(batch, args))
        else:
            obj, grad = fun(x, *chain(batch, args))
            objs.append(obj)

        norms.append(np.linalg.norm(grad))

        # Truncate gradients if bounded
        if bounds is not None:
            xlower = x <= lower
            grad[xlower] = np.minimum(grad[xlower], 0)
            xupper = x >= upper
            grad[xupper] = np.maximum(grad[xupper], 0)

        # perform update
        x = updater(x, grad)

        # Trucate steps if bounded
        if bounds is not None:
            x = np.clip(x, lower, upper)

    # Format results
    res = OptimizeResult(
        x=x,
        norms=norms,
        message='maxiter reached',
        fun=obj,
        objs=objs
    )

    return res


def gen_batch(data, batch_size, maxiter=np.inf, random_state=None):
    """
    Create random batches for Stochastic gradients.

    Batch index generator for SGD that will yeild random batches for a
    a defined number of iterations, which can be infinite. This generator makes
    consecutive passes through the data, drawing without replacement on each
    pass.

    Parameters
    ----------
    data : ndarray or sequence of ndarrays
        The data, can be a matrix X, (X,y) tuples etc
    batch_size : int
        number of data points in each batch.
    maxiter : int, optional
        The number of iterations
    random_state : int or RandomState, optional
        random seed

    Yields
    ------
    ndarray or sequence :
        with each array length ``batch_size``, i.e. a subset of data.
    """
    perms = endless_permutations(_len_data(data), random_state)

    it = 0
    while it < maxiter:
        it += 1
        ind = np.array([next(perms) for _ in range(batch_size)])
        yield _split_data(data, ind)


def normalize_bound(bound):
    """
    Replace ``None`` with + or - inf in bound tuples.

    Examples
    --------
    >>> normalize_bound((2.6, 7.2))
    (2.6, 7.2)

    >>> normalize_bound((None, 7.2))
    (-inf, 7.2)

    >>> normalize_bound((2.6, None))
    (2.6, inf)

    >>> normalize_bound((None, None))
    (-inf, inf)

    This operation is idempotent:

    >>> normalize_bound((-float("inf"), float("inf")))
    (-inf, inf)
    """
    min_, max_ = bound

    if min_ is None:
        min_ = -float('inf')

    if max_ is None:
        max_ = float('inf')

    return min_, max_


#
# Module Helpers
#

def _len_data(data):

    if not issequence(data):
        return data.shape[0]

    N = len(data[0])
    for d in data[1:]:
        if d.shape[0] != N:
            raise ValueError("Not all data is the same length!")

    return N


def _split_data(data, ind):

    if not issequence(data):
        return (data[ind],)

    return [d[ind] for d in data]



from .slm import StandardLinearModel
from .glm import GeneralizedLinearModel, GeneralisedLinearModel
from .btypes import Bound, Positive, Parameter
from . import likelihoods, basis_functions, metrics, btypes



"""
Reusable decorators
"""

from ..utils.base import flatten, unflatten
from collections import OrderedDict
from six import wraps

import numpy as np


class Memoize(dict):
    """
    Examples
    --------
    >>> @Memoize
    ... def fib(n):
    ...     if n < 2:
    ...         return n
    ...     return fib(n-2) + fib(n-1)

    >>> fib(10)
    55

    >>> isinstance(fib, dict)
    True

    >>> fib == {
    ...     (0,):  0,
    ...     (1,):  1,
    ...     (2,):  1,
    ...     (3,):  2,
    ...     (4,):  3,
    ...     (5,):  5,
    ...     (6,):  8,
    ...     (7,):  13,
    ...     (8,):  21,
    ...     (9,):  34,
    ...     (10,): 55,
    ... }
    True

    Order is not necessarily maintained.

    >>> sorted(fib.keys())
    [(0,), (1,), (2,), (3,), (4,), (5,), (6,), (7,), (8,), (9,), (10,)]

    >>> sorted(fib.values())
    [0, 1, 1, 2, 3, 5, 8, 13, 21, 34, 55]
    """
    def __init__(self, func):
        self.func = func
        super(Memoize, self).__init__()

    def __call__(self, *args):
        return self[args]

    def __missing__(self, key):
        self[key] = self.func(*key)
        return self[key]


class OrderedMemoize(Memoize, OrderedDict):
    """
    Examples
    --------
    >>> @OrderedMemoize
    ... def fib(n):
    ...     if n < 2:
    ...         return n
    ...     return fib(n-2) + fib(n-1)

    >>> fib(10)
    55

    The arguments and values are cached in the order they were called.

    >>> list(fib.keys())
    [(0,), (1,), (2,), (3,), (4,), (5,), (6,), (7,), (8,), (9,), (10,)]

    >>> list(fib.values())
    [0, 1, 1, 2, 3, 5, 8, 13, 21, 34, 55]

    >>> fib # doctest: +NORMALIZE_WHITESPACE
    OrderedMemoize([((0,), 0),
                    ((1,), 1),
                    ((2,), 1),
                    ((3,), 2),
                    ((4,), 3),
                    ((5,), 5),
                    ((6,), 8),
                    ((7,), 13),
                    ((8,), 21),
                    ((9,), 34),
                    ((10,), 55)])
    """
    pass


def flatten_args(fn):
    """
    Examples
    --------
    >>> @flatten_args
    ... def f(x):
    ...     return 2*x

    >>> x, y, z = f(np.array([1., 2.]), 3., np.array([[1., 2.],[.5, .9]]))

    >>> x
    array([ 2.,  4.])

    >>> y
    6.0

    >>> z
    array([[ 2. ,  4. ],
           [ 1. ,  1.8]])
    """
    @wraps(fn)
    def new_fn(*args):
        args_flat, shapes = flatten(args)
        result = fn(args_flat)
        return unflatten(result, shapes=shapes)

    return new_fn


def vectorize_args(fn):
    """
    When defining functions of several variables, it is usually more 
    readable to write out each variable as a separate argument. This is 
    also convenient for evaluating functions on a `numpy.meshgrid`. 

    However, the family of optimizers in `scipy.optimize` expects that 
    all functions, including those of several variables, receive a 
    single argument, which is a `numpy.ndarray` in the case of functions
    of several variables.

    Readability counts. We need not compromise readability to conform to
    some interface when higher-order functions/decorators can abstract 
    away the details for us. This is what this decorator does. 

    See Also
    --------
    revrand.utils.decorators.unvectorize_args

    Examples
    --------

    Optimizers such as those in `scipy.optimize` expects a function 
    defined like this.

    >>> def fun1(v):
    ...     # elliptic parabaloid
    ...     return 2*v[0]**2 + 2*v[1]**2 - 4
    
    >>> a = np.array([2, 3])
    
    >>> fun1(a)
    22

    Whereas this representation is not only more readable but more 
    natural.

    >>> def fun2(x, y):
    ...     # elliptic parabaloid
    ...     return 2*x**2 + 2*y**2 - 4
    
    >>> fun2(2, 3)
    22

    It is also important for evaluating functions on a `numpy.meshgrid`

    >>> y, x = np.mgrid[-5:5:0.2, -5:5:0.2]
    >>> fun2(x, y)
    array([[ 96.  ,  92.08,  88.32, ...,  84.72,  88.32,  92.08],
           [ 92.08,  88.16,  84.4 , ...,  80.8 ,  84.4 ,  88.16],
           [ 88.32,  84.4 ,  80.64, ...,  77.04,  80.64,  84.4 ],
           ..., 
           [ 84.72,  80.8 ,  77.04, ...,  73.44,  77.04,  80.8 ],
           [ 88.32,  84.4 ,  80.64, ...,  77.04,  80.64,  84.4 ],
           [ 92.08,  88.16,  84.4 , ...,  80.8 ,  84.4 ,  88.16]])

    We can easily reconcile the differences between these representation
    without having to compromise readability.

    >>> fun1(a) == vectorize_args(fun2)(a)
    True

    >>> @vectorize_args
    ... def fun3(x, y):
    ...     # elliptic parabaloid
    ...     return 2*x**2 + 2*y**2 - 4
    
    >>> fun1(a) == fun3(a)
    True
    """
    @wraps(fn)
    def new_fn(vec):
        return fn(*vec)
    return new_fn

def unvectorize_args(fn):

    """
    See Also
    --------
    revrand.utils.decorators.vectorize_args

    Examples
    --------
    The Rosenbrock function is commonly used as a performance test 
    problem for optimization algorithms. It and its derivatives are 
    included in `scipy.optimize` and is implemented as expected by the 
    family of optimization methods in `scipy.optimize`.

        def rosen(x):
            return sum(100.0*(x[1:]-x[:-1]**2.0)**2.0 + (1-x[:-1])**2.0)

    This representation makes it unwieldy to perform operations such as 
    plotting since it is less straightforward to evaluate the function 
    on a `meshgrid`. This decorator helps reconcile the differences 
    between these representations.

    >>> from scipy.optimize import rosen

    >>> rosen(np.array([0.5, 1.5]))
    156.5

    >>> unvectorize_args(rosen)(0.5, 1.5) 
    ... # doctest: +NORMALIZE_WHITESPACE
    156.5    

    The `rosen` function is implemented in such a way that it 
    generalizes to the Rosenbrock function of any number of variables. 
    This decorator supports can support any functions defined in a 
    similar manner.

    The function with any number of arguments are well-defined:

    >>> rosen(np.array([0.5, 1.5, 1., 0., 0.2]))
    418.0

    >>> unvectorize_args(rosen)(0.5, 1.5, 1., 0., 0.2)
    ... # can accept any variable number of arguments!
    418.0

    Make it easier to work with for other operations

    >>> rosen_ = unvectorize_args(rosen)
    >>> y, x = np.mgrid[0:2.1:0.05, -1:1.2:0.05]
    >>> z = rosen_(x, y)
    >>> z.round(2)
    array([[ 104.  ,   85.25,   69.22, ...,  121.55,  146.42,  174.92],
           [  94.25,   76.48,   61.37, ...,  110.78,  134.57,  161.95],
           [  85.  ,   68.2 ,   54.02, ...,  100.5 ,  123.22,  149.47],
           ..., 
           [  94.25,  113.53,  133.57, ...,   71.83,   54.77,   39.4 ],
           [ 104.  ,  124.25,  145.22, ...,   80.55,   62.42,   45.92],
           [ 114.25,  135.48,  157.37, ...,   89.78,   70.57,   52.95]])

    Now this can be directly plotted with `mpl_toolkits.mplot3d.Axes3D` 
    and `ax.plot_surface`.

    """
    @wraps(fn)
    def new_fn(*args):
        return fn(np.asarray(args))
    return new_fn

def vectorize_result(fn):
    @wraps(fn)
    def new_fn(*args):
        return np.asarray(fn(*args))
    return new_fn



""" Some validation functions. """

from __future__ import division

import numpy as np
from scipy.stats import norm


def smse(y_true, y_pred):
    """
    Standardised mean squared error.

    Parameters
    ----------
    y_true: ndarray
        vector of true targets
    y_pred: ndarray
        vector of predicted targets

    Returns
    -------
    float:
        SMSE of predictions vs truth (scalar)

    Example
    -------
    >>> y_true = np.random.randn(100)
    >>> smse(y_true, y_true)
    0.0
    >>> smse(y_true, np.random.randn(100)) >= 1.0
    True
    """

    N = y_true.shape[0]
    return ((y_true - y_pred)**2).sum() / (N * y_true.var())


def mll(y_true, y_pred, y_var):
    """
    Mean log loss under a Gaussian distribution.

    Parameters
    ----------
    y_true: ndarray
        vector of true targets
    y_pred: ndarray
        vector of predicted targets
    y_var: float or ndarray
        predicted variances

    Returns
    -------
    float:
        The mean negative log loss (negative log likelihood)

    Example
    -------
    >>> y_true = np.random.randn(100)
    >>> mean_prob = - norm.logpdf(1e-2, loc=0)  # -ve log prob close to mean
    >>> mll(y_true, y_true, 1) <= mean_prob  # Should be good predictor
    True
    >>> mll(y_true, np.random.randn(100), 1) >= mean_prob  # naive predictor
    True
    """

    return - norm.logpdf(y_true, loc=y_pred, scale=np.sqrt(y_var)).mean()


def msll(y_true, y_pred, y_var, y_train):
    """
    Mean standardised log loss under a Gaussian distribution.

    Parameters
    ----------
    y_true: ndarray
        vector of true targets
    y_pred: ndarray
        vector of predicted targets
    y_var: float or ndarray
        predicted variances
    y_train: ndarray
        vector of *training* targets by which to standardise

    Returns
    -------
    float:
        The negative mean standardised log loss (negative log likelihood)

    Example
    -------
    >>> y_true = np.random.randn(100)
    >>> msll(y_true, y_true, 1, y_true) < 0  # Should be good predictor
    True
    >>> msll(y_true, np.random.randn(100), 1, y_true) >= 0  # naive predictor
    True
    """

    var = y_train.var()
    mu = y_train.mean()

    ll_naive = norm.logpdf(y_true, loc=mu, scale=np.sqrt(var))
    ll_mod = norm.logpdf(y_true, loc=y_pred, scale=np.sqrt(y_var))

    return - (ll_mod - ll_naive).mean()


def lins_ccc(y_true, y_pred):
    """
    Lin's Concordance Correlation Coefficient.

    See https://en.wikipedia.org/wiki/Concordance_correlation_coefficient

    Parameters
    ----------
    y_true: ndarray
        vector of true targets
    y_pred: ndarray
        vector of predicted targets

    Returns
    -------
    float:
        1.0 for a perfect match between :code:`y_true` and :code:`y_pred`, less
        otherwise

    Example
    -------
    >>> y_true = np.random.randn(100)
    >>> lins_ccc(y_true, y_true) > 0.99  # Should be good predictor
    True
    >>> lins_ccc(y_true, np.zeros_like(y_true)) < 0.01  # Bad predictor
    True
    """

    t = y_true.mean()
    p = y_pred.mean()
    St = y_true.var()
    Sp = y_pred.var()
    Spt = np.mean((y_true - t) * (y_pred - p))

    return 2 * Spt / (St + Sp + (t - p)**2)



"""Likelihood objects for inference within the GLM framework."""

from __future__ import division

import numpy as np

from scipy.stats import bernoulli, binom, poisson, norm, gamma
from scipy.special import gammaln, expit

from .btypes import Parameter, Positive
from .mathfun.special import safesoftplus, softplus


#
# Likelihood classes
#

class Bernoulli():
    r"""
    Bernoulli likelihood class for (binary) classification tasks.

    A logistic transformation function is used to map the latent function from
    the GLM prior into a probability.

    .. math::

        p(y_i | f_i) = \sigma(f_i) ^ {y_i} (1 - \sigma(f_i))^{1 - y_i}

    where :math:`y_i` is a target, :math:`f_i` the value of the latent function
    corresponding to the target, and :math:`\sigma(\cdot)` is the logistic
    sigmoid.
    """

    _params = Parameter()

    @property
    def params(self):
        """Get this object's Parameter types."""
        return self._params

    @params.setter
    def params(self, params):
        """Set this object's Parameter types."""
        self._params = params

    def loglike(self, y, f):
        r"""
        Bernoulli log likelihood.

        Parameters
        ----------
        y: ndarray
            array of 0, 1 valued integers of targets
        f: ndarray
            latent function from the GLM prior (:math:`\mathbf{f} =
            \boldsymbol\Phi \mathbf{w}`)

        Returns
        -------
        logp: ndarray
            the log likelihood of each y given each f under this
            likelihood.
        """
        # way faster than calling bernoulli.logpmf
        y, f = np.broadcast_arrays(y, f)
        ll = y * f - softplus(f)
        return ll

    def Ey(self, f):
        r"""
        Expected value of the Bernoulli likelihood.

        Parameters
        ----------
        f: ndarray
            latent function from the GLM prior (:math:`\mathbf{f} =
            \boldsymbol\Phi \mathbf{w}`)

        Returns
        -------
        Ey: ndarray
            expected value of y, :math:`\mathbb{E}[\mathbf{y}|\mathbf{f}]`.
        """
        return expit(f)

    def df(self, y, f):
        r"""
        Derivative of Bernoulli log likelihood w.r.t.\  f.

        Parameters
        ----------
        y: ndarray
            array of 0, 1 valued integers of targets
        f: ndarray
            latent function from the GLM prior (:math:`\mathbf{f} =
            \boldsymbol\Phi \mathbf{w}`)

        Returns
        -------
        df: ndarray
            the derivative :math:`\partial \log p(y|f) / \partial f`
        """
        y, f = np.broadcast_arrays(y, f)
        return y - expit(f)

    def dp(self, y, f, *args):
        r"""
        Derivative of Bernoulli log likelihood w.r.t.\ the parameters,
        :math:`\theta`.

        Parameters
        ----------
        y: ndarray
            array of 0, 1 valued integers of targets
        f: ndarray
            latent function from the GLM prior (:math:`\mathbf{f} =
            \boldsymbol\Phi \mathbf{w}`)

        Returns
        -------
        dp: list, float or ndarray
            the derivative
            :math:`\partial \log p(y|f, \theta)/ \partial \theta` for
            each parameter. If there is only one parameter, this is not a
            list.
        """
        return []

    def cdf(self, y, f):
        r"""
        Cumulative density function of the likelihood.

        Parameters
        ----------
        y: ndarray
            query quantiles, i.e.\  :math:`P(Y \leq y)`.
        f: ndarray
            latent function from the GLM prior (:math:`\mathbf{f} =
            \boldsymbol\Phi \mathbf{w}`)

        Returns
        -------
        cdf: ndarray
            Cumulative density function evaluated at y.
        """
        return bernoulli.cdf(y, expit(f))

    def __repr__(self):
        """Representation."""
        return "{}()".format(self.__class__.__name__)


class Binomial(Bernoulli):
    r"""
    Binomial likelihood class.

    A logistic transformation function is used to map the latent function from
    the GLM prior into a probability.

    .. math::

        p(y_i | f_i) = \genfrac(){0pt}{}{n}{y_i}
            \sigma(f_i) ^ {y_i} (1 - \sigma(f_i))^{n - y_i}

    where :math:`y_i` is a target, :math:`f_i` the value of the latent function
    corresponding to the target, :math:`n` is the total possible count, and
    :math:`\sigma(\cdot)` is the logistic sigmoid. :math:`n` can also be
    applied per observation.
    """

    def loglike(self, y, f, n):
        r"""
        Binomial log likelihood.

        Parameters
        ----------
        y: ndarray
            array of 0, 1 valued integers of targets
        f: ndarray
            latent function from the GLM prior (:math:`\mathbf{f} =
            \boldsymbol\Phi \mathbf{w}`)
        n: ndarray
            the total number of observations

        Returns
        -------
        logp: ndarray
            the log likelihood of each y given each f under this
            likelihood.
        """
        ll = binom.logpmf(y, n=n, p=expit(f))
        return ll

    def Ey(self, f, n):
        r"""
        Expected value of the Binomial likelihood.

        Parameters
        ----------
        f: ndarray
            latent function from the GLM prior (:math:`\mathbf{f} =
            \boldsymbol\Phi \mathbf{w}`)
        n: ndarray
            the total number of observations

        Returns
        -------
        Ey: ndarray
            expected value of y, :math:`\mathbb{E}[\mathbf{y}|\mathbf{f}]`.
        """
        return expit(f) * n

    def df(self, y, f, n):
        r"""
        Derivative of Binomial log likelihood w.r.t.\  f.

        Parameters
        ----------
        y: ndarray
            array of 0, 1 valued integers of targets
        f: ndarray
            latent function from the GLM prior (:math:`\mathbf{f} =
            \boldsymbol\Phi \mathbf{w}`)
        n: ndarray
            the total number of observations

        Returns
        -------
        df: ndarray
            the derivative :math:`\partial \log p(y|f) / \partial f`
        """
        y, f, n = np.broadcast_arrays(y, f, n)
        return y - expit(f) * n

    def cdf(self, y, f, n):
        r"""
        Cumulative density function of the likelihood.

        Parameters
        ----------
        y: ndarray
            query quantiles, i.e.\  :math:`P(Y \leq y)`.
        f: ndarray
            latent function from the GLM prior (:math:`\mathbf{f} =
            \boldsymbol\Phi \mathbf{w}`)
        n: ndarray
            the total number of observations

        Returns
        -------
        cdf: ndarray
            Cumulative density function evaluated at y.
        """
        return binom.cdf(y, n=n, p=expit(f))

    def __repr__(self):
        """Representation."""
        return "{}()".format(self.__class__.__name__)


class Gaussian(Bernoulli):
    r"""
    A univariate Gaussian likelihood for general regression tasks.

    No transformation function is needed since this is (conditionally)
    conjugate to the GLM prior.

    .. math::

        p(y_i | f_i) = \frac{1}{\sqrt{2 \pi \sigma^2}}
            \exp\left(- \frac{(y_i - f_i)^2}{2 \sigma^2} \right)

    where :math:`y_i` is a target, :math:`f_i` the value of the latent function
    corresponding to the target and :math:`\sigma` is the observation noise
    (standard deviation).

    Parameters
    ----------
    var: Parameter, optional
        A scalar Parameter describing the initial point and bounds for
        an optimiser to learn the variance parameter of this object.
    """

    def __init__(self, var=Parameter(gamma(1., scale=1), Positive())):
        """See class docstring."""
        self.params = var

    def _check_param(self, param):

        if param is None:
            return self.params.value

        if not self.params.bounds.check(param):
            raise ValueError("Input parameter is out of bounds!")

        return param

    def loglike(self, y, f, var=None):
        r"""
        Gaussian log likelihood.

        Parameters
        ----------
        y: ndarray
            array of 0, 1 valued integers of targets
        f: ndarray
            latent function from the GLM prior (:math:`\mathbf{f} =
            \boldsymbol\Phi \mathbf{w}`)
        var: float, ndarray, optional
            The variance of the distribution, if not input, the initial value
            of variance is used.

        Returns
        -------
        logp: ndarray
            the log likelihood of each y given each f under this
            likelihood.
        """
        # way faster than calling norm.logpdf
        var = self._check_param(var)
        y, f = np.broadcast_arrays(y, f)
        ll = - 0.5 * (np.log(2 * np.pi * var) + (y - f)**2 / var)
        return ll

    def Ey(self, f, var):
        r"""
        Expected value of the Gaussian likelihood.

        Parameters
        ----------
        f: ndarray
            latent function from the GLM prior (:math:`\mathbf{f} =
            \boldsymbol\Phi \mathbf{w}`)
        var: float, ndarray, optional
            The variance of the distribution, if not input, the initial value
            of variance is used.

        Returns
        -------
        Ey: ndarray
            expected value of y, :math:`\mathbb{E}[\mathbf{y}|\mathbf{f}]`.
        """
        var = self._check_param(var)
        return f

    def df(self, y, f, var):
        r"""
        Derivative of Gaussian log likelihood w.r.t.\  f.

        Parameters
        ----------
        y: ndarray
            array of 0, 1 valued integers of targets
        f: ndarray
            latent function from the GLM prior (:math:`\mathbf{f} =
            \boldsymbol\Phi \mathbf{w}`)
        var: float, ndarray, optional
            The variance of the distribution, if not input, the initial value
            of variance is used.

        Returns
        -------
        df: ndarray
            the derivative :math:`\partial \log p(y|f) / \partial f`
        """
        var = self._check_param(var)
        y, f = np.broadcast_arrays(y, f)
        return (y - f) / var

    def dp(self, y, f, var):
        r"""
        Derivative of Gaussian log likelihood w.r.t.\ the variance
        :math:`\sigma^2`.

        Parameters
        ----------
        y: ndarray
            array of 0, 1 valued integers of targets
        f: ndarray
            latent function from the GLM prior (:math:`\mathbf{f} =
            \boldsymbol\Phi \mathbf{w}`)
        var: float, ndarray, optional
            The variance of the distribution, if not input, the initial value
            of variance is used.

        Returns
        -------
        dp: float
            the derivative
            :math:`\partial \log p(y|f, \sigma^2)/ \partial \sigma^2`
            where :math:`sigma^2` is the variance.
        """
        var = self._check_param(var)
        y, f = np.broadcast_arrays(y, f)
        ivar = 1. / var
        return 0.5 * (((y - f) * ivar)**2 - ivar)

    def cdf(self, y, f, var):
        r"""
        Cumulative density function of the likelihood.

        Parameters
        ----------
        y: ndarray
            query quantiles, i.e.\  :math:`P(Y \leq y)`.
        f: ndarray
            latent function from the GLM prior (:math:`\mathbf{f} =
            \boldsymbol\Phi \mathbf{w}`)
        var: float, ndarray, optional
            The variance of the distribution, if not input, the initial value
            of variance is used.

        Returns
        -------
        cdf: ndarray
            Cumulative density function evaluated at y.
        """
        var = self._check_param(var)
        return norm.cdf(y, loc=f, scale=np.sqrt(var))

    def __repr__(self):
        """Representation."""
        return "{}(var={})".format(self.__class__.__name__, self.params)


class Poisson(Bernoulli):
    r"""
    A Poisson likelihood, useful for various Poisson process tasks.

    An exponential transformation function and a softplus transformation
    function have been implemented.

    .. math::

        p(y_i | f_i) = \frac{g(f_i)^{y_i} e^{-g(f_i)}}{y_i!}

    where :math:`y_i` is a target, :math:`f_i` the value of the latent function
    corresponding to the target, and :math:`g(\cdot)` is the tranformation
    function, which can be either an exponential function, or a softplus
    function (:math:`\log(1 + \exp(f_i)`).

    Parameters
    ----------
    tranfcn: string, optional
        this may be 'exp' for an exponential transformation function,
        or 'softplus' for a softplut transformation function.
    """

    def __init__(self, tranfcn='exp'):
        """See class docstring."""
        if tranfcn == 'exp' or tranfcn == 'softplus':
            self.tranfcn = tranfcn
        else:
            raise ValueError('Invalid transformation function specified!')

    def loglike(self, y, f):
        r"""
        Poisson log likelihood.

        Parameters
        ----------
        y: ndarray
            array of integer targets
        f: ndarray
            latent function from the GLM prior (:math:`\mathbf{f} =
            \boldsymbol\Phi \mathbf{w}`)

        Returns
        -------
        logp: ndarray
            the log likelihood of each y given each f under this
            likelihood.
        """
        y, f = np.broadcast_arrays(y, f)
        if self.tranfcn == 'exp':
            g = np.exp(f)
            logg = f
        else:
            g = softplus(f)
            logg = np.log(g)
        return y * logg - g - gammaln(y + 1)

    def Ey(self, f):
        r"""
        Expected value of the Poisson likelihood.

        Parameters
        ----------
        f: ndarray
            latent function from the GLM prior (:math:`\mathbf{f} =
            \boldsymbol\Phi \mathbf{w}`)

        Returns
        -------
        Ey: ndarray
            expected value of y, :math:`\mathbb{E}[\mathbf{y}|\mathbf{f}]`.
        """
        return np.exp(f) if self.tranfcn == 'exp' else softplus(f)

    def df(self, y, f):
        r"""
        Derivative of Poisson log likelihood w.r.t.\  f.

        Parameters
        ----------
        y: ndarray
            array of 0, 1 valued integers of targets
        f: ndarray
            latent function from the GLM prior (:math:`\mathbf{f} =
            \boldsymbol\Phi \mathbf{w}`)

        Returns
        -------
        df: ndarray
            the derivative :math:`\partial \log p(y|f) / \partial f`
        """
        y, f = np.broadcast_arrays(y, f)
        if self.tranfcn == 'exp':
            return y - np.exp(f)
        else:
            return expit(f) * (y / safesoftplus(f) - 1)

    def cdf(self, y, f):
        r"""
        Cumulative density function of the likelihood.

        Parameters
        ----------
        y: ndarray
            query quantiles, i.e.\  :math:`P(Y \leq y)`.
        f: ndarray
            latent function from the GLM prior (:math:`\mathbf{f} =
            \boldsymbol\Phi \mathbf{w}`)

        Returns
        -------
        cdf: ndarray
            Cumulative density function evaluated at y.
        """
        mu = np.exp(f) if self.tranfcn == 'exp' else softplus(f)
        return poisson.cdf(y, mu=mu)

    def __repr__(self):
        """Representation."""
        return "{}(tranfcn='{}')".format(self.__class__.__name__, self.tranfcn)



"""
Various basis function objects specialised for parameter learning.

To make a new basis object, see the documentation of the Basis class.

"""

from __future__ import division

import sys
import inspect
from itertools import repeat
from functools import reduce

import numpy as np
from six import wraps
from decorator import decorator  # Preserves function signature (pyth2 compat)
from scipy.linalg import norm
from scipy.stats import gamma, norm as norm_dist
from scipy.special import expit
from scipy.spatial.distance import cdist
from sklearn.utils import check_random_state

from .btypes import Positive, Bound, Parameter
from .mathfun.linalg import hadamard
from .utils import issequence, atleast_list, atleast_tuple


#
# Module Helper Functions
#

# For basis concatenation functionality
if sys.version_info[0] < 3:
    def count_args(func):
        """
        Count the number of arguments in a function/method.

        Parameters
        ----------
        func: callable
            a function or class method

        Returns
        -------
        int:
            the number of arguments, excluding self
        """
        nargs = len(inspect.getargspec(func)[0])
        return nargs - 1 if inspect.ismethod(func) else nargs  # remove self
else:
    def count_args(func):
        """
        Count the number of arguments in a function/method.

        Parameters
        ----------
        func: callable
            a function or class method

        Returns
        -------
        int:
            the number of arguments, excluding self
        """
        return len((inspect.signature(func)).parameters)


# For basis function slicing
def slice_init(func):
    """
    Decorator for adding partial application functionality to a basis object.

    This will add an "apply_ind" argument to a basis object initialiser that
    can be used to apply the basis function to only the dimensions specified in
    apply_ind. E.g.,

    >>> X = np.ones((100, 20))
    >>> base = LinearBasis(onescol=False, apply_ind=slice(0, 10))
    >>> base.transform(X).shape
    (100, 10)
    """
    @wraps(func)
    def new_init(self, *args, **kwargs):

        apply_ind = kwargs.pop('apply_ind', None)
        if np.isscalar(apply_ind):
            apply_ind = [apply_ind]

        func(self, *args, **kwargs)
        self.apply_ind = apply_ind

    return new_init


@decorator  # This needs to be signature preserving for concatenation
def slice_transform(func, self, X, *vargs, **kwargs):
    """
    Decorator for implementing partial application.

    This must decorate the ``transform`` and ``grad`` methods of basis objects
    if the ``slice_init`` decorator was used.
    """
    X = X if self.apply_ind is None else X[:, self.apply_ind]
    return func(self, X, *vargs, **kwargs)


# Calculating function gradients w.r.t. structured basis functions
def apply_grad(fun, grad):
    """
    Apply a function that takes a gradient matrix to a sequence of 2 or 3
    dimensional gradients.

    This is partucularly useful when the gradient of a basis concatenation
    object is quite complex, eg.

    >>> X = np.random.randn(100, 3)
    >>> y = np.random.randn(100)
    >>> N, d = X.shape
    >>> base = RandomRBF(Xdim=d, nbases=5) + RandomRBF(Xdim=d, nbases=5,
    ... lenscale=Parameter(np.ones(d), Positive()))
    >>> Phi = base.transform(X, 1., np.ones(d))
    >>> dffun = lambda dPhi: y.dot(Phi).dot(dPhi.T).dot(y)
    >>> df = apply_grad(dffun, base.grad(X, 1., np.ones(d)))
    >>> np.isscalar(df[0])
    True
    >>> df[1].shape
    (3,)

    Parameters
    ----------
    fun: callable
        the function too apply to the (2d) gradient.
    grad: ndarray or generator
        the gradient of the basis function (output of base.grad).

    Returns
    -------
    scalar, ndarray or sequence:
        the result of applying fun(grad) for a structured grad.
    """
    if issequence(grad):
        fgrad = [apply_grad(fun, g) for g in grad]
        return fgrad if len(fgrad) != 1 else fgrad[0]
    elif len(grad) == 0:
        return []
    elif (grad.ndim == 1) or (grad.ndim == 2):
        return fun(grad)
    elif grad.ndim == 3:
        return np.array([fun(grad[:, :, i]) for i in range(grad.shape[2])])
    else:
        raise ValueError("Only up to 3d gradients allowed!")


#
# Basis objects
#

class Basis(object):
    """
    The base Basis class.

    To make other basis classes, make sure they are subclasses of this class to
    enable concatenation and operation with the machine learning algorithms.

    Parameters
    ----------
    regularizer : None, Parameter, optional
        The (initial) value of the regularizer/prior variance to apply to the
        regression weights of this basis function. The Parameter object must
        have a scalar value. If it is not set, it will take on a default value
        of ``Parameter(gamma(1.), Positive())``.

    Example
    -------
    Basis concatentation works as follows if you subclass this class:

    >>> base = MyBasis1(properties1) + MyBasis2(properties2)  # doctest: +SKIP
    """

    _params = Parameter()
    _regularizer = Parameter(gamma(1.), Positive())

    @slice_init
    def __init__(self, regularizer=None):
        r"""
        Construct this an instance of this class.

        This is also a good place to set non-learnable properties, and bounded
        Parameter types. An example Basis class with parameters may be

        Example
        -------
        >>> def __init__(self, # doctest: +SKIP
        ...              property,
        ...              param=Parameter(1, Bound()),
        ...              regularizer=Parameter(gamma(1.), Positive())
        ...              ):
        ...
        ...     self.property = property
        ...     self._params = param_init
        ...     super().__init__(regularizer)

        All the ``params`` property does is inform algorithms of the intitial
        value and any bounds this basis object has, or it can be used as a
        default value. This will need to correspond to any parameters input
        into the ``transform`` and ``grad`` methods, where it can be used as
        the default value. All basis class objects MUST have a ``params``
        property, which is either:

        - one Parameter object for an optimisable parameter, see btypes.py.
          Parameter objects with ``None`` values are interpreted as having no
          parameters.
        - a list of ``Parameter`` objects, one for each optimisable parameter.

        Also, all basis classes can have a way of setting the
        ``self.regularizer`` property, since this informs the regression
        algorithms of the (initial) values of the weights regularizer
        corresponding to this basis. If it is not set, it will take on a
        default value of ``Parameter(gamma(1.), Positive())``.
        """
        if regularizer is not None:
            if not regularizer.is_scalar:
                raise ValueError("Regularizer parameters have to be scalar!")
            if regularizer.bounds.lower <= 0:
                raise ValueError("Regularizer has to be bounded below by 0!")
            self._regularizer = regularizer

    @slice_transform
    def transform(self, X):
        """
        Return the basis function applied to X.

        I.e. Phi(X, params), where params can also optionally be used and
        learned.

        Parameters
        ----------
        X : ndarray
            (N, d) array of observations where N is the number of samples, and
            d is the dimensionality of X.
        params : optional
            parameter aguments, these can be scalars or arrays.

        Returns
        -------
        ndarray :
            of shape (N, D) where D is the number of basis functions.
        """
        return X

    @slice_transform
    def grad(self, X):
        """
        Return the gradient of the basis function for each parameter.

        Parameters
        ----------
        X : ndarray
            (N, d) array of observations where N is the number of samples, and
            d is the dimensionality of X.
        params : optional
            parameter aguments, these can be scalars or arrays.

        Returns
        -------
        list or ndarray :
            this will be a list of ndarrays if there are multiple parameters,
            or just an ndarray if there is a single parameter. The ndarrays can
            have more than two dimensions (i.e. tensors of rank > 2), depending
            on the dimensions of the basis function parameters. If there are
            *no* parameters, ``[]`` is returned.
        """
        return []

    def get_dim(self, X):
        """
        Get the output dimensionality of this basis.

        This makes a cheap call to transform with the initial parameter values
        to ascertain the dimensionality of the output features.

        Parameters
        ----------
        X : ndarray
            (N, d) array of observations where N is the number of samples, and
            d is the dimensionality of X.

        Returns
        -------
        int :
            The dimensionality of the basis.
        """
        # Cache
        if not hasattr(self, '_D'):
            self._D = self.transform(X[[0]], *self.params_values()).shape[1]
        return self._D

    def params_values(self):
        """
        Get a list of the ``Parameter`` values if they have a value.

        This does not include the basis regularizer.
        """
        return [p.value for p in atleast_list(self.params) if p.has_value]

    def regularizer_diagonal(self, X, regularizer=None):
        r"""
        Get the diagonal of the prior variance on the weights (regularizer).

        This also returns a slice into the regression weights to which this
        regularizer applies.

        Parameters
        ----------
        X : ndarray
            (N, d) array of observations where N is the number of samples, and
            d is the dimensionality of X.
        regularizer : float, optional
            use this value instead of the value passed into the constructor for
            the regularizer.

        Returns
        -------
        ndarray :
            an array of shape (D,) that is the diagonal of the prior variance
            of the weights, :math:`\boldsymbol\Lambda` in :math:`\mathbf{w}
            \sim \mathcal{N}(\mathbf{0}, \boldsymbol\Lambda)`.
        slice:
            the slice into the regression weights that this regularizer
            applies to.

        Note
        ----
        You should not have to modify this method if you inherit from ``Basis``
        unless you are doing something very interesting...
        """
        reg = self.regularizer.value if regularizer is None else regularizer
        diag = np.full(self.get_dim(X), reg, dtype=float)
        return diag, slice(None)

    def _transform_popargs(self, X, *args):

        selfargs, otherargs = self.__splitargs(args, self.transform, offset=1)

        return self.transform(X, *selfargs), otherargs

    def _grad_popargs(self, X, *args):

        selfargs, otherargs = self.__splitargs(args, self.grad, offset=1)

        return self.grad(X, *selfargs), otherargs, selfargs

    def __splitargs(self, args, fn, offset=0):

        nargs = count_args(fn) - offset
        selfargs, otherargs = args[:nargs], args[nargs:]

        return selfargs, otherargs

    @property
    def params(self):
        """Get this basis' Parameter types."""
        return self._params

    @property
    def regularizer(self):
        """Get the ``Parameter`` value of this basis' regularizer."""
        return self._regularizer

    def __add__(self, other):
        """Concatenation."""
        return BasisCat([self, other])

    def __radd__(self, other):
        """Concatenation."""
        return self if other == 0 else self.__add__(other)

    def __repr__(self):
        """Representation."""
        return "{}(regularizer={})".format(
            self.__class__.__name__,
            self.regularizer
        )


class BiasBasis(Basis):
    r"""
    Bias Basis for adding a bias term to a regressor.

    This just returns a column of a constant value so a bias term can be
    learned by a regressor.

    .. math::

        \phi(\mathbf{X}) = \mathbf{1} \times \text{const}

    Parameters
    ----------
    offset: float, optional
        A scalar value to give the bias column. By default this is one.
    regularizer : None, Parameter, optional
        The (initial) value of the regularizer/prior variance to apply to the
        regression weights of this basis function. The Parameter object must
        have a scalar value. If it is not set, it will take on a default value
        of ``Parameter(gamma(1.), Positive())``.
    """

    @slice_init
    def __init__(self, offset=1., regularizer=None):
        """See this class's docstring."""
        self.offset = offset
        super(BiasBasis, self).__init__(regularizer)

    @slice_transform
    def transform(self, X):
        """
        Return this basis applied to X.

        Parameters
        ----------
        X: ndarray
            of shape (N, d) of observations where N is the number of samples,
            and d is the dimensionality of X.

        Returns
        -------
        ndarray:
            of shape (N, 1) of ones * self.offset.
        """
        N = len(X)
        return np.ones((N, 1)) * self.offset

    def __repr__(self):
        """Representation."""
        return "{}(offset={}, regularizer={})".format(
            self.__class__.__name__,
            self.offset,
            self.regularizer
        )


class LinearBasis(Basis):
    r"""
    Linear basis class, basically this just prepends a columns of ones onto X.

    .. math::

        \phi(\mathbf{X}) = [\mathbf{1}, \mathbf{X}]

    Parameters
    ----------
    onescol: bool, optional
        If true, prepend a column of ones onto X.
    regularizer : None, Parameter, optional
        The (initial) value of the regularizer/prior variance to apply to the
        regression weights of this basis function. The Parameter object must
        have a scalar value. If it is not set, it will take on a default value
        of ``Parameter(gamma(1.), Positive())``.
    """

    @slice_init
    def __init__(self, onescol=False, regularizer=None):
        """See this class's docstring."""
        self.onescol = onescol
        super(LinearBasis, self).__init__(regularizer)

    @slice_transform
    def transform(self, X):
        """
        Return this basis applied to X.

        Parameters
        ----------
        X: ndarray
            of shape (N, d) of observations where N is the number of samples,
            and d is the dimensionality of X.

        Returns
        -------
        ndarray:
            of shape (N, d+1), or (N, d) depending on onescol.
        """
        N, D = X.shape
        return np.hstack((np.ones((N, 1)), X)) if self.onescol else X

    def __repr__(self):
        """Representation."""
        return "{}(onescol={}, regularizer={})".format(
            self.__class__.__name__,
            self.onescol,
            self.regularizer
        )


class PolynomialBasis(Basis):
    r"""
    Polynomial basis class.

    This essentially creates the concatenation,

    .. math::

        \phi(\mathbf{X}) = [\mathbf{1}, \mathbf{X}^1, \ldots, \mathbf{X}^p]

    where :math:`p` is the :code:`order` of the polynomial.

    Parameters
    ----------
    order: int
        the order of the polynomial to create.
    include_bias: bool, optional
        If True (default), include the bias column (column of ones which
        acts as the intercept term in a linear model)
    regularizer : None, Parameter, optional
        The (initial) value of the regularizer/prior variance to apply to the
        regression weights of this basis function. The Parameter object must
        have a scalar value. If it is not set, it will take on a default value
        of ``Parameter(gamma(1.), Positive())``.
    """

    @slice_init
    def __init__(self, order, include_bias=True, regularizer=None):
        """See this class's docstring."""
        if order < 0:
            raise ValueError("Polynomial order must be positive")
        self.order = order

        self.include_bias = include_bias
        super(PolynomialBasis, self).__init__(regularizer)

    @slice_transform
    def transform(self, X):
        """
        Return this basis applied to X.

        Parameters
        ----------
        X: ndarray
            of shape (N, d) of observations where N is the number of samples,
            and d is the dimensionality of X.

        Returns
        -------
        ndarray:
            of shape (N, d*order+1), the extra 1 is from a prepended ones
            column.
        """
        N, D = X.shape

        pow_arr = np.arange(self.order) + 1

        # Polynomial terms
        Phi = X[:, :, np.newaxis] ** pow_arr

        # Flatten along last axes
        Phi = Phi.reshape(N, D * self.order)

        # Prepend intercept
        if self.include_bias:
            Phi = np.hstack((np.ones((N, 1)), Phi))

        # TODO: Using np.hstack is about 4x slower than initializing, say,
        # an N by d*order+1 ndarray of ones and assigning the remaining
        # N by d*order values. May want to revisit this implementation.

        return Phi

    def __repr__(self):
        """Representation."""
        return "{}(order={}, include_bias={}, regularizer={})".format(
            self.__class__.__name__,
            self.order,
            self.include_bias,
            self.regularizer
        )


class _LengthScaleBasis(Basis):

    def _init_lenscale(self, lenscale):

        if (lenscale.shape != (self.d,)) \
                and (lenscale.shape != ()):
            raise ValueError("Parameter dimension doesn't agree with X"
                             " dimensions!")

        self._params = lenscale

    def _check_dim(self, Xdim, in_param, paramind=None):

        if Xdim != self.d:
            raise ValueError("Dimensions of data inconsistent!")

        # Get the right internal parameter for comparison
        sparam = self.params if paramind is None else self.params[paramind]

        # If param was not input, get intitial value
        if in_param is None:
            in_param = sparam.value

        # Check input parameter is within bounds
        sparam.bounds.check(in_param)

        # Promote dimension of parameter
        if np.isscalar(in_param):
            in_param = np.array([in_param])

        if (sparam.shape == () and len(in_param) == 1) \
                or np.shape(in_param) == sparam.shape:
            return in_param
        else:
            raise ValueError("Dimension of input parameter is inconsistent!")


class RadialBasis(_LengthScaleBasis):
    r"""
    Radial basis class.

    .. math::

        \phi(\mathbf{X}) =
            \exp \left( -\frac{\|\mathbf{X} - \mathbf{C}\|^2} {2 l^2} \right)

    Where :math:`\mathbf{C}` are radial basis centres, and :math:`l` is a
    length scale.

    Parameters
    ----------
    centres: ndarray
        array of shape (Dxd) where D is the number of centres for the
        radial bases, and d is the dimensionality of X.
    lenscale: Parameter, optional
        A scalar or vector of shape (1,) or (d,) Parameter to bound and
        initialise the length scales for optimization. If this is shape (d,),
        ARD length scales will be expected, otherwise an isotropic lenscale is
        learned.
    regularizer : None, Parameter, optional
        The (initial) value of the regularizer/prior variance to apply to the
        regression weights of this basis function. The Parameter object must
        have a scalar value. If it is not set, it will take on a default value
        of ``Parameter(gamma(1.), Positive())``.

    Note
    ----
    This will have relevance vector machine-like behaviour with uncertainty.
    """

    @slice_init
    def __init__(
            self,
            centres,
            lenscale=Parameter(gamma(1.), Positive()),
            regularizer=None
    ):
        """See this class's docstring."""
        self.M, self.d = centres.shape
        self.C = centres
        self._init_lenscale(lenscale)
        super(_LengthScaleBasis, self).__init__(regularizer)

    @slice_transform
    def transform(self, X, lenscale=None):
        """
        Apply the RBF to X.

        Parameters
        ----------
        X: ndarray
            (N, d) array of observations where N is the number of samples, and
            d is the dimensionality of X.
        lenscale: scalar or ndarray, optional
            scalar or array of shape (d,) length scales (one for each dimension
            of X). If not input, this uses the value of the initial length
            scale.

        Returns
        -------
        ndarray:
            of shape (N, D) where D is number of RBF centres.
        """
        N, d = X.shape
        lenscale = self._check_dim(d, lenscale)

        den = (2 * lenscale**2)
        return np.exp(- cdist(X / den, self.C / den, 'sqeuclidean'))

    @slice_transform
    def grad(self, X, lenscale=None):
        r"""
        Get the gradients of this basis w.r.t.\ the length scale.

        Parameters
        ----------
        X: ndarray
            (N, d) array of observations where N is the number of samples, and
            d is the dimensionality of X.
        lenscale: scalar or ndarray, optional
            scalar or array of shape (d,) length scales (one for each dimension
            of X). If not input, this uses the value of the initial length
            scale.

        Returns
        -------
        ndarray:
            of shape (N, D) where D is number of RBF centres. This is
            :math:`\partial \Phi(\mathbf{X}) / \partial l`
        """
        N, d = X.shape
        lenscale = self._check_dim(d, lenscale)

        Phi = self.transform(X, lenscale)
        dPhi = []
        for i, l in enumerate(lenscale):
            ldist = cdist(X[:, [i]] / l**3, self.C[:, [i]] / l**3,
                          'sqeuclidean')
            dPhi.append(Phi * ldist)

        return np.dstack(dPhi) if len(lenscale) != 1 else dPhi[0]

    def __repr__(self):
        """Representation."""
        return "{}(centres={}, lenscale={}, regularizer={})".format(
            self.__class__.__name__,
            self.C,
            self.params,
            self.regularizer
        )


class SigmoidalBasis(RadialBasis):
    r"""
    Sigmoidal Basis.

    .. math::

        \phi(\mathbf{X}) =
            \sigma \left( \frac{\|\mathbf{X} - \mathbf{C}\|}{l} \right)

    where :math:`\mathbf{C}` are sigmoidal basis centres, :math:`l` is a
    length scale and :math:`\sigma` is the logistic sigmoid function defined by

    .. math::

        \sigma(a) = \frac{1}{1+e^{-a}}.

    Parameters
    ----------
    centres: ndarray
        array of shape (Dxd) where D is the number of centres for the bases,
        and d is the dimensionality of X.
    lenscale: Parameter, optional
        A scalar parameter to bound and initialise the length scales for
        optimization.
    regularizer : None, Parameter, optional
        The (initial) value of the regularizer/prior variance to apply to the
        regression weights of this basis function. The Parameter object must
        have a scalar value. If it is not set, it will take on a default value
        of ``Parameter(gamma(1.), Positive())``.
    """

    @slice_transform
    def transform(self, X, lenscale=None):
        r"""
        Apply the sigmoid basis function to X.

        Parameters
        ----------
        X: ndarray
            (N, d) array of observations where N is the number of samples, and
            d is the dimensionality of X.
        lenscale: float
            the length scale (scalar) of the RBFs to apply to X. If not input,
            this uses the value of the initial length scale.

        Returns
        -------
        ndarray:
            of shape (N, D) where D is number of centres.
        """
        N, d = X.shape
        lenscale = self._check_dim(d, lenscale)

        return expit(cdist(X / lenscale, self.C / lenscale, 'euclidean'))

    @slice_transform
    def grad(self, X, lenscale=None):
        r"""
        Get the gradients of this basis w.r.t.\  the length scale.

        Parameters
        ----------
        X: ndarray
            (N, d) array of observations where N is the number of samples, and
            d is the dimensionality of X.
        lenscale: float, optional
            the length scale (scalar) of the RBFs to apply to X. If not input,
            this uses the value of the initial length scale.

        Returns
        -------
        ndarray:
            of shape (N, D) where D is number of centres. This is
            :math:`\partial \Phi(\mathbf{X}) / \partial l`
        """
        N, d = X.shape
        lenscale = self._check_dim(d, lenscale)

        Phi = self.transform(X, lenscale)
        dPhi = []
        for i, l in enumerate(lenscale):
            ldist = cdist(X[:, [i]] / l**2, self.C[:, [i]] / l**2, 'euclidean')
            dPhi.append(- ldist * Phi * (1 - Phi))

        return np.dstack(dPhi) if len(lenscale) != 1 else dPhi[0]


class _RandomKernelBasis(_LengthScaleBasis):
    """Base class for the random kernel approximation bases."""

    @slice_init
    def __init__(self,
                 nbases,
                 Xdim,
                 lenscale=Parameter(gamma(1.), Positive()),
                 regularizer=None,
                 random_state=None
                 ):
        """See this class's docstring."""
        self.d = Xdim
        self.n = nbases
        self.random_state = random_state  # for repr
        self._random = check_random_state(random_state)
        self.W = self._weightsamples()
        self._init_lenscale(lenscale)
        super(_LengthScaleBasis, self).__init__(regularizer)

    @slice_transform
    def transform(self, X, lenscale=None):
        """
        Apply the random basis to X.

        Parameters
        ----------
        X: ndarray
            (N, d) array of observations where N is the number of samples, and
            d is the dimensionality of X.
        lenscale: scalar or ndarray, optional
            scalar or array of shape (d,) length scales (one for each dimension
            of X). If not input, this uses the value of the initial length
            scale.

        Returns
        -------
        ndarray:
            of shape (N, 2*nbases) where nbases is number of random bases to
            use, given in the constructor.
        """
        N, D = X.shape
        lenscale = self._check_dim(D, lenscale)[:, np.newaxis]

        WX = np.dot(X, self.W / lenscale)

        return np.hstack((np.cos(WX), np.sin(WX))) / np.sqrt(self.n)

    @slice_transform
    def grad(self, X, lenscale=None):
        r"""
        Get the gradients of this basis w.r.t.\ the length scales.

        Parameters
        ----------
        X: ndarray
            (N, d) array of observations where N is the number of samples, and
            d is the dimensionality of X.
        lenscale: scalar or ndarray, optional
            scalar or array of shape (d,) length scales (one for each dimension
            of X). If not input, this uses the value of the initial length
            scale.

        Returns
        -------
        ndarray:
            of shape (N, 2*nbases[, d]) where d is number of lenscales (if not
            ARD, i.e. scalar lenscale, this is just a 2D array). This is
            :math:`\partial \Phi(\mathbf{X}) / \partial \mathbf{l}`
        """
        N, D = X.shape
        lenscale = self._check_dim(D, lenscale)[:, np.newaxis]

        WX = np.dot(X, self.W / lenscale)
        sinWX = - np.sin(WX)
        cosWX = np.cos(WX)

        dPhi = []
        for i, l in enumerate(lenscale):
            dWX = np.outer(X[:, i], - self.W[i, :] / l**2)
            dPhi.append(np.hstack((dWX * sinWX, dWX * cosWX)) /
                        np.sqrt(self.n))

        return np.dstack(dPhi) if len(lenscale) != 1 else dPhi[0]

    def __repr__(self):
        """Representation."""
        return "{}(nbases={}, Xdim={}, lenscale={}, regularizer={}," \
            " random_state={})".format(
                self.__class__.__name__,
                self.n,
                self.d,
                self.params,
                self.regularizer,
                self.random_state
            )


class RandomRBF(_RandomKernelBasis):
    r"""
    Random RBF Basis -- Approximates an RBF kernel function.

    This will make a linear regression model approximate a GP with an
    (optionally ARD) RBF covariance function,

    .. math::

        \phi(\mathbf{x})^\top \phi(\mathbf{x}') \approx
            \exp\left( -\frac{\| \mathbf{x} - \mathbf{x}' \|^2}{2 l^2} \right)

    with a length scale, :math:`l` (a vector in :math:`\mathbb{R}^d` for ARD).

    Parameters
    ----------
    nbases: int
        how many unique random bases to create (twice this number will be
        actually created, i.e. real and imaginary components for each base)
    Xdim: int
        the dimension (d) of the observations (or the dimension of the slices
        if using apply_ind).
    lenscale: Parameter, optional
        A scalar or vector of shape (1,) or (d,) Parameter to bound and
        initialise the length scales for optimization. If this is shape (d,),
        ARD length scales will be expected, otherwise an isotropic lenscale is
        learned.
    regularizer : None, Parameter, optional
        The (initial) value of the regularizer/prior variance to apply to the
        regression weights of this basis function. The Parameter object must
        have a scalar value. If it is not set, it will take on a default value
        of ``Parameter(gamma(1.), Positive())``.
    random_state: None, int or RandomState, optional
        random seed
    """

    def _weightsamples(self):
        weights = self._random.randn(self.d, self.n)
        return weights


class RandomLaplace(_RandomKernelBasis):
    r"""
    Random Laplace Basis -- Approximates a Laplace kernel function.

    This will make a linear regression model approximate a GP with an
    (optionally ARD) Laplace covariance function.

    .. math::

        \phi(\mathbf{x})^\top \phi(\mathbf{x}') \approx
            \exp\left( -\frac{\| \mathbf{x} - \mathbf{x}' \|}{l} \right)

    with a length scale, :math:`l` (a vector in :math:`\mathbb{R}^d` for ARD).

    Parameters
    ----------
    nbases: int
        how many unique random bases to create (twice this number will be
        actually created, i.e. real and imaginary components for each base)
    Xdim: int
        the dimension (d) of the observations (or the dimension of the slices
        if using apply_ind).
    lenscale: Parameter, optional
        A scalar or vector of shape (1,) or (d,) Parameter to bound and
        initialise the length scales for optimization. If this is shape (d,),
        ARD length scales will be expected, otherwise an isotropic lenscale is
        learned.
    regularizer : None, Parameter, optional
        The (initial) value of the regularizer/prior variance to apply to the
        regression weights of this basis function. The Parameter object must
        have a scalar value. If it is not set, it will take on a default value
        of ``Parameter(gamma(1.), Positive())``.
    random_state: None, int or RandomState, optional
        random seed
    """

    def _weightsamples(self):
        weights = self._random.standard_cauchy(size=(self.d, self.n))
        return weights


class RandomCauchy(_RandomKernelBasis):
    r"""
    Random Cauchy Basis -- Approximates a Cauchy kernel function.

    This will make a linear regression model approximate a GP with an
    (optionally ARD) Cauchy covariance function.

    .. math::

        \phi(\mathbf{x})^\top \phi(\mathbf{x}') \approx
            \frac{1}{1 + (\| \mathbf{x} - \mathbf{x}' \| / l)^2}

    with a length scale, :math:`l` (a vector in :math:`\mathbb{R}^d` for ARD).

    Parameters
    ----------
    nbases: int
        how many unique random bases to create (twice this number will be
        actually created, i.e. real and imaginary components for each base)
    Xdim: int
        the dimension (d) of the observations (or the dimension of the slices
        if using apply_ind).
    lenscale: Parameter, optional
        A scalar or vector of shape (1,) or (d,) Parameter to bound and
        initialise the length scales for optimization. If this is shape (d,),
        ARD length scales will be expected, otherwise an isotropic lenscale is
        learned.
    regularizer : None, Parameter, optional
        The (initial) value of the regularizer/prior variance to apply to the
        regression weights of this basis function. The Parameter object must
        have a scalar value. If it is not set, it will take on a default value
        of ``Parameter(gamma(1.), Positive())``.
    random_state: None, int or RandomState, optional
        random seed
    """

    def _weightsamples(self):

        # A draw from a (regular) mv laplace is the same as:
        # X ~ Norm(mu, cov)
        # Z ~ gamma(1)
        # Y ~ (2 * Z) * X
        # See "Multivariate Generalized Laplace Distributions and Related
        # Random Fields":
        #   http://www.math.chalmers.se/Math/Research/Preprints/2010/47.pdf
        X = self._random.randn(self.d, self.n)
        Z = self._random.standard_gamma(1., size=(1, self.n))
        return X * np.sqrt(2 * Z)


class _RandomMatern(_RandomKernelBasis):
    """Base slass for random Matern kernel basis approximations."""

    def _maternweight(self, p):

        # p is the matern number (v = p + .5) and the two is a transformation
        # of variables between Rasmussen 2006 p84 and the CF of a Multivariate
        # Student t (see wikipedia). Also see "A Note on the Characteristic
        # Function of Multivariate t Distribution":
        #   http://ocean.kisti.re.kr/downfile/volume/kss/GCGHC8/2014/v21n1/
        #   GCGHC8_2014_v21n1_81.pdf
        # To sample from a m.v. t we use the formula
        # from wikipedia, x = y * np.sqrt(df / u) where y ~ norm(0, I),
        # u ~ chi2(df), then x ~ mvt(0, I, df)
        df = 2 * (p + 0.5)
        y = self._random.randn(self.d, self.n)
        u = self._random.chisquare(df, size=(self.n,))
        return y * np.sqrt(df / u)


class RandomMatern32(_RandomMatern):
    r"""
    Random Matern 3/2 Basis -- Approximates a Matern 3/2 kernel function.

    This will make a linear regression model approximate a GP with an
    (optionally ARD) Matern covariance function.

    .. math::

        \phi(\mathbf{x})^\top \phi(\mathbf{x}') \approx
            \left(1 + \frac{\sqrt{3} \| \mathbf{x} - \mathbf{x}' \|}{l} \right)
            \exp
            \left(- \frac{\sqrt{3} \| \mathbf{x} - \mathbf{x}' \|}{l} \right)

    with a length scale, :math:`l` (a vector in :math:`\mathbb{R}^d` for ARD).

    Parameters
    ----------
    nbases: int
        how many unique random bases to create (twice this number will be
        actually created, i.e. real and imaginary components for each base)
    Xdim: int
        the dimension (d) of the observations (or the dimension of the slices
        if using apply_ind).
    lenscale: Parameter, optional
        A scalar or vector of shape (1,) or (d,) Parameter to bound and
        initialise the length scales for optimization. If this is shape (d,),
        ARD length scales will be expected, otherwise an isotropic lenscale is
        learned.
    regularizer : None, Parameter, optional
        The (initial) value of the regularizer/prior variance to apply to the
        regression weights of this basis function. The Parameter object must
        have a scalar value. If it is not set, it will take on a default value
        of ``Parameter(gamma(1.), Positive())``.
    random_state: None, int or RandomState, optional
        random seed
    """

    def _weightsamples(self):
        return self._maternweight(p=1)


class RandomMatern52(_RandomMatern):
    r"""
    Random Matern 5/2 Basis -- Approximates a Matern 5/2 kernel function.

    This will make a linear regression model approximate a GP with an
    (optionally ARD) Matern covariance function.

    .. math::

        \phi(\mathbf{x})^\top \phi(\mathbf{x}') \approx
            \left(1 + \frac{\sqrt{5} \| \mathbf{x} - \mathbf{x}' \|}{l}
                + \frac{5 \| \mathbf{x} - \mathbf{x}' \|^2}{3l^2} \right)
            \exp
            \left(- \frac{\sqrt{5} \| \mathbf{x} - \mathbf{x}' \|}{l} \right)

    with a length scale, :math:`l` (a vector in :math:`\mathbb{R}^d` for ARD).

    Parameters
    ----------
    nbases: int
        how many unique random bases to create (twice this number will be
        actually created, i.e. real and imaginary components for each base)
    Xdim: int
        the dimension (d) of the observations (or the dimension of the slices
        if using apply_ind).
    lenscale: Parameter, optional
        A scalar or vector of shape (1,) or (d,) Parameter to bound and
        initialise the length scales for optimization. If this is shape (d,),
        ARD length scales will be expected, otherwise an isotropic lenscale is
        learned.
    regularizer : None, Parameter, optional
        The (initial) value of the regularizer/prior variance to apply to the
        regression weights of this basis function. The Parameter object must
        have a scalar value. If it is not set, it will take on a default value
        of ``Parameter(gamma(1.), Positive())``.
    random_state: None, int or RandomState, optional
        random seed
    """

    def _weightsamples(self):
        return self._maternweight(p=2)


class FastFoodRBF(_LengthScaleBasis):
    r"""
    Fast Food radial basis function.

    This is an approximation of the random radial basis function for a large
    number of bases.

    .. math::

        \phi(\mathbf{x})^\top \phi(\mathbf{x}') \approx
            \exp\left( -\frac{\| \mathbf{x} - \mathbf{x}' \|^2}{2 l^2} \right)

    with a length scale, :math:`l` (a vector in :math:`\mathbb{R}^d` for ARD).

    Parameters
    ----------
    nbases: int
        a scalar for how many (unique) random bases to create approximately,
        this actually will be to the nearest larger two power.
    Xdim: int
        the dimension (d) of the observations (or the dimension of the slices
        if using apply_ind).
    lenscale: Parameter, optional
        A scalar or vector of shape (1,) or (d,) Parameter to bound and
        initialise the length scales for optimization. If this is shape (d,),
        ARD length scales will be expected, otherwise an isotropic lenscale is
        learned.
    regularizer : None, Parameter, optional
        The (initial) value of the regularizer/prior variance to apply to the
        regression weights of this basis function. The Parameter object must
        have a scalar value. If it is not set, it will take on a default value
        of ``Parameter(gamma(1.), Positive())``.
    random_state: None, int or RandomState, optional
        random seed
    """

    @slice_init
    def __init__(self,
                 nbases,
                 Xdim,
                 lenscale=Parameter(gamma(1.), Positive()),
                 regularizer=None,
                 random_state=None
                 ):
        """See this class's docstring."""
        self.random_state = random_state  # for repr
        self._random = check_random_state(random_state)
        self._init_dims(nbases, Xdim)
        self._init_lenscale(lenscale)
        self._init_matrices()
        super(_LengthScaleBasis, self).__init__(regularizer)

    @slice_transform
    def transform(self, X, lenscale=None):
        """
        Apply the Fast Food RBF basis to X.

        Parameters
        ----------
        X: ndarray
            (N, d) array of observations where N is the number of samples, and
            d is the dimensionality of X.
        lenscale: scalar or ndarray, optional
            scalar or array of shape (d,) length scales (one for each dimension
            of X).If not input, this uses the value of the initial length
            scale.


        Returns
        -------
        ndarray:
            of shape (N, 2*nbases) where nbases is number of random bases to
            use, given in the constructor (to nearest larger two power).
        """
        lenscale = self._check_dim(X.shape[1], lenscale)

        VX = self._makeVX(X / lenscale)
        Phi = np.hstack((np.cos(VX), np.sin(VX))) / np.sqrt(self.n)
        return Phi

    @slice_transform
    def grad(self, X, lenscale=None):
        r"""
        Get the gradients of this basis w.r.t.\ the length scale.

        Parameters
        ----------
        x: ndarray
            (n, d) array of observations where n is the number of samples, and
            d is the dimensionality of x.
        lenscale: scalar or ndarray
            scalar or array of shape (d,) length scales (one for each dimension
            of x).If not input, this uses the value of the initial length
            scale.


        Returns
        -------
        ndarray:
            shape (n, 2*nbases) where nbases is number of random rbf bases,
            again to the nearest larger two power. This is
            :math:`\partial \phi(\mathbf{x}) / \partial l`
        """
        d = X.shape[1]
        lenscale = self._check_dim(d, lenscale)

        VX = self._makeVX(X / lenscale)
        sinVX = - np.sin(VX)
        cosVX = np.cos(VX)

        dPhi = []
        for i, l in enumerate(lenscale):
            indlen = np.zeros(d)
            indlen[i] = 1. / l**2
            dVX = - self._makeVX(X * indlen)  # FIXME make this more efficient?
            dPhi.append(np.hstack((dVX * sinVX, dVX * cosVX)) /
                        np.sqrt(self.n))

        return np.dstack(dPhi) if len(lenscale) != 1 else dPhi[0]

    def _init_dims(self, nbases, Xdim):

        # Make sure our dimensions are powers of 2
        l = int(np.ceil(np.log2(Xdim)))

        self.nbases = nbases
        self.d = Xdim
        self.d2 = pow(2, l)
        self.k = int(np.ceil(nbases / self.d2))
        self.n = self.d2 * self.k

    def _init_matrices(self):

        # Draw consistent samples from the covariance matrix
        shape = (self.k, self.d2)
        self.B = self._random.randint(2, size=shape) * 2 - 1  # uniform [-1,1]
        self.G = self._random.randn(*shape)  # mean 0 std 1
        self.PI = np.array([self._random.permutation(self.d2)
                            for _ in range(self.k)])
        self.S = self._weightsamples()

    def _weightsamples(self):
        s = np.sqrt(self._random.chisquare(self.d2, size=self.G.shape))
        return self.d2 * s / norm(self.G, axis=1)[:, np.newaxis]

    def _makeVX(self, X):
        N, d0 = X.shape

        # Pad the dimensions of X to nearest 2 power
        X_dash = np.zeros((N, self.d2))
        X_dash[:, 0:d0] = X

        VX = []
        for B, G, PI, S in zip(*(self.B, self.G, self.PI, self.S)):
            vX = hadamard(X_dash * B[np.newaxis, :], ordering=False)
            vX = vX[:, PI] * G[np.newaxis, :]
            vX = hadamard(vX, ordering=False) * S[np.newaxis, :] * \
                np.sqrt(self.d2)
            VX.append(vX)

        return np.hstack(VX)

    def __repr__(self):
        """Representation."""
        return "{}(nbases={}, Xdim={}, lenscale={}, regularizer={}," \
            " random_state={})".format(
                self.__class__.__name__,
                self.nbases,
                self.d,
                self.params,
                self.regularizer,
                self.random_state
            )


class FastFoodGM(FastFoodRBF):
    """
    A mixture component from a Gaussian spectral mixture kernel approximation.

    This implements a GM basis component from "A la Carte - Learning Fast
    Kernels". This essentially learns the form of a kernel function, and so has
    no explicit kernel representation!

    To fully implement a Gaussian spectral mixture, concatenate as many of
    these objects as desired (one per mixture component). Also remember to
    initialize all of the bases with different means.

    Parameters
    ----------
    nbases: int
        a scalar for how many (unique) random bases to create approximately,
        this actually will be to the nearest larger two power.
    Xdim: int
        the dimension (d) of the observations (or the dimension of the slices
        if using apply_ind).
    mean: Parameter, optional
        A scalar or vector of shape (1,) or (d,) Parameter to bound and
        initialise the component frequency means for optimization. This will
        always initialise (d,) means if a scalr bound is given, it is applied
        to all means.
    lenscale: Parameter, optional
        A scalar or vector of shape (1,) or (d,) Parameter to bound and
        initialise the length scales for optimization. This will always
        initialise ARD length scales, if a scalr bound is given, it is applied
        to all length scales.
    regularizer : None, Parameter, optional
        The (initial) value of the regularizer/prior variance to apply to the
        regression weights of this basis function. The Parameter object must
        have a scalar value. If it is not set, it will take on a default value
        of ``Parameter(gamma(1.), Positive())``.
    random_state: None, int or RandomState, optional
        random seed
    """

    @slice_init
    def __init__(self,
                 nbases,
                 Xdim,
                 mean=Parameter(norm_dist(), Bound()),
                 lenscale=Parameter(gamma(1.), Positive()),
                 regularizer=None,
                 random_state=None
                 ):
        """See this class's docstring."""
        self.random_state = random_state  # for repr
        self._random = check_random_state(random_state)
        self._init_dims(nbases, Xdim)
        self._params = [self._init_param(mean),
                        self._init_param(lenscale)]
        self._init_matrices()
        super(_LengthScaleBasis, self).__init__(regularizer)

    @slice_transform
    def transform(self, X, mean=None, lenscale=None):
        """
        Apply the spectral mixture component basis to X.

        Parameters
        ----------
        X: ndarray
            (N, d) array of observations where N is the number of samples, and
            d is the dimensionality of X.
        mean: ndarray, optional
            array of shape (d,) frequency means (one for each dimension of X).
            If not input, this uses the value of the initial mean.
        lenscale: ndarray, optional
            array of shape (d,) length scales (one for each dimension of X). If
            not input, this uses the value of the initial length scale.

        Returns
        -------
        ndarray:
            of shape (N, 4*nbases) where nbases is number of random bases to
            use, given in the constructor (to nearest larger two power).
        """
        mean = self._check_dim(X.shape[1], mean, paramind=0)
        lenscale = self._check_dim(X.shape[1], lenscale, paramind=1)

        VX = self._makeVX(X / lenscale)
        mX = X.dot(mean)[:, np.newaxis]
        Phi = np.hstack((np.cos(VX + mX), np.sin(VX + mX),
                         np.cos(VX - mX), np.sin(VX - mX))) / \
            np.sqrt(2 * self.n)

        return Phi

    @slice_transform
    def grad(self, X, mean=None, lenscale=None):
        r"""
        Get the gradients of this basis w.r.t.\ the mean and length scales.

        Parameters
        ----------
        x: ndarray
            (n, d) array of observations where n is the number of samples, and
            d is the dimensionality of x.
        mean: ndarray, optional
            array of shape (d,) frequency means (one for each dimension of X).
            If not input, this uses the value of the initial mean.
        lenscale: ndarray, optional
            array of shape (d,) length scales (one for each dimension of X). If
            not input, this uses the value of the initial length scale.

        Returns
        -------
        ndarray:
            shape (n, 4*nbases) where nbases is number of random rbf bases,
            again to the nearest larger two power. This is
            :math:`\partial \phi(\mathbf{x}) / \partial \boldsymbol\mu`
        ndarray:
            shape (n, 4*nbases) where nbases is number of random rbf bases,
            again to the nearest larger two power. This is
            :math:`\partial \phi(\mathbf{x}) / \partial \mathbf{l}`
        """
        d = X.shape[1]
        mean = self._check_dim(d, mean, paramind=0)
        lenscale = self._check_dim(d, lenscale, paramind=1)

        VX = self._makeVX(X / lenscale)
        mX = X.dot(mean)[:, np.newaxis]

        sinVXpmX = - np.sin(VX + mX)
        sinVXmmX = - np.sin(VX - mX)
        cosVXpmX = np.cos(VX + mX)
        cosVXmmX = np.cos(VX - mX)

        dPhi_len = []
        dPhi_mean = []
        for i, l in enumerate(lenscale):

            # Means
            dmX = X[:, [i]]
            dPhi_mean.append(np.hstack((dmX * sinVXpmX, dmX * cosVXpmX,
                                        -dmX * sinVXmmX, -dmX * cosVXmmX)) /
                             np.sqrt(2 * self.n))

            # Lenscales
            indlen = np.zeros(d)
            indlen[i] = 1. / l**2
            dVX = - self._makeVX(X * indlen)  # FIXME make this more efficient?
            dPhi_len.append(np.hstack((dVX * sinVXpmX, dVX * cosVXpmX,
                                       dVX * sinVXmmX, dVX * cosVXmmX)) /
                            np.sqrt(2 * self.n))

        dPhi_mean = np.dstack(dPhi_mean) if d != 1 else dPhi_mean[0]
        dPhi_len = np.dstack(dPhi_len) if d != 1 else dPhi_len[0]
        return dPhi_mean, dPhi_len

    def _init_param(self, param):

        if param.shape == (self.d,):
            return param
        elif param.shape in ((), (1,)):
            param.shape = (self.d,)
            param.value = np.ones(self.d) * param.value
            return param
        else:
            raise ValueError("Parameter dimension doesn't agree with X"
                             " dimensions!")

    def __repr__(self):
        """Representation."""
        return "{}(nbases={}, Xdim={}, mean={}, lenscale={}, regularizer={}," \
            " random_state={})".format(
                self.__class__.__name__,
                self.nbases,
                self.d,
                self.params[0],
                self.params[1],
                self.regularizer,
                self.random_state
            )


#
# Other basis construction objects and functions
#

class BasisCat(object):
    """
    A class that implements concatenation of bases.

    Parameters
    ----------
    basis_list : list
        a list of ``Basis`` objects to concatenate
    """

    def __init__(self, basis_list):
        """See the class docstring for details."""
        # Concatenate BasisCat lists not objects here
        def check_BasisCat(blist, b):
            rlist = atleast_list(blist)

            if isinstance(b, BasisCat):
                rlist.extend(b.bases)
            else:
                rlist.append(b)

            return rlist

        self.bases = reduce(check_BasisCat, basis_list)

        # Caches
        self.__dims = None
        self.__baseinds = None
        self.__slices = None

    def transform(self, X, *params):
        """
        Return the basis function applied to X.

        I.e. Phi(X, params), where params can also optionally be used and
        learned.

        Parameters
        ----------
        X : ndarray
            (N, d) array of observations where N is the number of samples, and
            d is the dimensionality of X.
        *params : optional
            parameter aguments, these are the parameters of the concatenated
            bases `in the order` they were concatenated.

        Returns
        -------
        ndarray :
            of shape (N, D) where D is the number of basis functions.
        """
        Phi = []
        args = list(params)

        for base in self.bases:
            phi, args = base._transform_popargs(X, *args)
            Phi.append(phi)

        return np.hstack(Phi)

    def grad(self, X, *params):
        """
        Return the gradient of the basis function for each parameter.

        Parameters
        ----------
        X : ndarray
            (N, d) array of observations where N is the number of samples, and
            d is the dimensionality of X.
        *params : optional
            parameter aguments, these are the parameters of the concatenated
            bases `in the order` they were concatenated.

        Returns
        -------
        list or ndarray :
            this will be a list of ndarrays if there are multiple parameters,
            or just an ndarray if there is a single parameter. The ndarrays can
            have more than two dimensions (i.e. tensors of rank > 2), depending
            on the dimensions of the basis function parameters. If there are
            *no* parameters, ``[]`` is returned.
        """
        # Establish a few dimensions
        N = X.shape[0]
        D = self.get_dim(X)
        endinds = self.__base_locations(X)  # for the Padding indices
        args = list(params)

        # Generate structured gradients with appropriate zero padding
        def make_dPhi(i, g):

            # Pad the gradient with respect to the total basis dimensionality
            dPhi_dim = (N, D) if g.ndim < 3 else (N, D, g.shape[2])
            dPhi = np.zeros(dPhi_dim)
            dPhi[:, endinds[i]:endinds[i + 1]] = g

            return dPhi

        # Get gradients from each basis
        for i, base in enumerate(self.bases):

            # evaluate gradient and deal with multiple parameter gradients by
            # keeping track of the basis index
            g, args, sargs = base._grad_popargs(X, *args)

            for gg in atleast_tuple(g):
                if len(gg) == 0:
                    continue
                yield make_dPhi(i, gg)

    def get_dim(self, X):
        """
        Get the output dimensionality of this basis.

        This makes a cheap call to transform with the initial parameter values
        to ascertain the dimensionality of the output features.

        Parameters
        ----------
        X : ndarray
            (N, d) array of observations where N is the number of samples, and
            d is the dimensionality of X.

        Returns
        -------
        int :
            The dimensionality of the basis.
        """
        return np.sum(self.__all_dims(X))

    def params_values(self):
        """
        Get a list of the ``Parameter`` values if they have a value.

        This does not include the basis regularizers.
        """
        return [v for b in self.bases for v in b.params_values()]

    @property
    def regularizer(self):
        """Return a list of concatenated bases regularizers."""
        return [b.regularizer for b in self.bases]

    def regularizer_diagonal(self, X, *regularizer):
        r"""
        Get the diagonal of the prior variance on the weights (regularizer).

        This also returns a slice into the regression weights to which all of
        the concatenated bases' regularizers applies.

        Parameters
        ----------
        X : ndarray
            (N, d) array of observations where N is the number of samples, and
            d is the dimensionality of X.
        *regularizer :
            use these values instead of the values passed into the ``Basis``
            constructors for each regularizer. These are applied to each basis
            in the concatentation in the order they were concatenated.

        Returns
        -------
        ndarray :
            an array of shape (D,) that is the diagonal of the prior variance
            of the weights, :math:`\boldsymbol\Lambda` in :math:`\mathbf{w}
            \sim \mathcal{N}(\mathbf{0}, \boldsymbol\Lambda)`.
        list of slices:
            the slices into the regression weights that each regularizer
            applies to.
        """
        regularizer = repeat(None) if regularizer is () else regularizer
        regs, _ = zip(*(b.regularizer_diagonal(X, r)
                        for b, r in zip(self.bases, regularizer)))

        if self.__slices is None:
            baseinds = self.__base_locations(X)
            self.__slices = [slice(b, e)
                             for b, e in zip(baseinds[:-1], baseinds[1:])]

        return np.concatenate(regs), self.__slices

    @property
    def params(self):
        """
        Return a list of all of the ``Parameter`` objects.

        Or a just a single ``Parameter`` is there is only one, and single empty
        ``Parameter`` if there are no parameters.
        """
        paramlist = [b.params for b in self.bases if b.params.has_value]

        if len(paramlist) == 0:
            return Parameter()
        else:
            return paramlist if len(paramlist) > 1 else paramlist[0]

    def __all_dims(self, X):

        if self.__dims is None:
            self.__dims = [b.get_dim(X) for b in self.bases]
        return self.__dims

    def __base_locations(self, X):

        if self.__baseinds is None:
            self.__baseinds = np.cumsum([0] + self.__all_dims(X))
        return self.__baseinds

    def __add__(self, other):
        """Concatenation."""
        if isinstance(other, BasisCat):
            return BasisCat(self.bases + other.bases)
        else:
            return BasisCat(self.bases + [other])

    def __radd__(self, other):
        """Concatenation."""
        return self if other == 0 else self.__add__(other)

    def __repr__(self):
        """Representation."""
        return "{}(basis_list={})".format(self.__class__.__name__, self.bases)



"""Optimizer Decorators."""
import logging
from functools import partial
from itertools import chain
from six import wraps

import numpy as np

import revrand.btypes as bt
from ..utils import flatten, unflatten, map_recursive
from .sgd import gen_batch

# Set up logging
log = logging.getLogger(__name__)


# Constants
MINPOS = 1e-100  # Min for log trick warped data
MAXPOS = np.sqrt(np.finfo(float).max)  # Max for log trick warped data
LOGMINPOS = np.log(MINPOS)
EXPMAX = np.log(MAXPOS)


def structured_minimizer(minimizer):
    r"""
    Allow an optimizer to accept nested sequences of Parameters to optimize.

    This decorator can intepret the :code:`Parameter` objects in `btypes.py`,
    and can accept nested sequences of *any* structure of these objects to
    optimise!

    It can also optionally evaluate *random starts* (i.e. random starting
    candidates) if the parameter objects have been initialised with
    distributions. For this, two additional parameters are exposed in the
    :code:`minimizer` interface.

    Parameters
    ----------
    fun : callable
        objective function that takes in arbitrary ndarrays, floats or nested
        sequences of these.
    parameters : (nested) sequences of Parameter objects
        Initial guess of the parameters of the objective function
    nstarts : int, optional
        The number random starting candidates for optimisation to evaluate.
        This will only happen for :code:`nstarts > 0` and if at least one
        :code:`Parameter` object is random.
    random_state : None, int or RandomState, optional
        random seed

    Examples
    --------
    >>> from scipy.optimize import minimize as sp_min
    >>> from ..btypes import Parameter, Bound

    Define a cost function that returns a pair. The first element is the cost
    value and the second element is the gradient represented by a tuple. Even
    if the cost is a function of a single variable, the gradient must be a
    tuple containing one element.

    >>> def cost(w, lambda_):
    ...     sq_norm = w.T.dot(w)
    ...     return .5 * lambda_ * sq_norm, (lambda_ * w, .5 * sq_norm)

    Augment the Scipy optimizer to take structured inputs

    >>> new_min = structured_minimizer(sp_min)

    Constant Initial values

    >>> w_0 = Parameter(np.array([.5, .1, .2]), Bound())
    >>> lambda_0 = Parameter(.25, Bound())

    >>> res = new_min(cost, (w_0, lambda_0), method='L-BFGS-B', jac=True)
    >>> res_w, res_lambda = res.x

    Random Initial values

    >>> from scipy.stats import norm, gamma
    >>> w_0 = Parameter(norm(), Bound(), shape=(3,))
    >>> lambda_0 = Parameter(gamma(a=1), Bound())

    >>> res = new_min(cost, (w_0, lambda_0), method='L-BFGS-B', jac=True,
    ...               nstarts=100, random_state=None)
    >>> res_w, res_lambda = res.x
    """
    @wraps(minimizer)
    def new_minimizer(fun, parameters, jac=True, args=(), nstarts=0,
                      random_state=None, **minimizer_kwargs):

        (array1d, fbounds), shapes = flatten(
            parameters,
            hstack=bt.hstack,
            shape=bt.shape,
            ravel=partial(bt.ravel, random_state=random_state)
        )

        # Find best random starting candidate if we are doing random starts
        if nstarts > 0:
            array1d = _random_starts(
                fun=fun,
                parameters=parameters,
                jac=jac,
                args=args,
                nstarts=nstarts,
                random_state=random_state
            )

        # Wrap function calls to work with wrapped minimizer
        flatten_args_dec = flatten_args(shapes)
        new_fun = flatten_args_dec(fun)

        # Wrap gradient calls to work with wrapped minimizer
        if callable(jac):
            new_jac = flatten_args_dec(jac)
        else:
            new_jac = jac
            if bool(jac):
                new_fun = flatten_func_grad(new_fun)

        result = minimizer(new_fun, array1d, jac=new_jac, args=args,
                           bounds=fbounds, **minimizer_kwargs)
        result['x'] = tuple(unflatten(result['x'], shapes))

        if bool(jac):
            result['jac'] = tuple(unflatten(result['jac'], shapes))

        return result

    return new_minimizer


def structured_sgd(sgd):
    r"""
    Allow an SGD to accept nested sequences of Parameters to optimize.

    This decorator can intepret the :code:`Parameter` objects in `btypes.py`,
    and can accept nested sequences of *any* structure of these objects to
    optimise!

    It can also optionally evaluate *random starts* (i.e. random starting
    candidates) if the parameter objects have been initialised with
    distributions. For this, an additional parameter is exposed in the
    :code:`minimizer` interface.

    Parameters
    ----------
    fun : callable
        objective function that takes in arbitrary ndarrays, floats or nested
        sequences of these.
    parameters : (nested) sequences of Parameter objects
        Initial guess of the parameters of the objective function
    nstarts : int, optional
        The number random starting candidates for optimisation to evaluate.
        This will only happen for :code:`nstarts > 0` and if at least one
        :code:`Parameter` object is random.

    Examples
    --------
    >>> from ..optimize import sgd
    >>> from ..btypes import Parameter, Bound

    Define a cost function that returns a pair. The first element is the cost
    value and the second element is the gradient represented by a sequence.
    Even if the cost is a function of a single variable, the gradient must be a
    sequence containing one element.

    >>> def cost(w, lambda_, data):
    ...     N = len(data)
    ...     y, X = data[:, 0], data[:, 1:]
    ...     y_est = X.dot(w)
    ...     ww = w.T.dot(w)
    ...     obj = (y - y_est).sum() / N + lambda_ * ww
    ...     gradw = - 2 * X.T.dot(y - y_est) / N + 2 * lambda_ * w
    ...     gradl = ww
    ...     return obj, [gradw, gradl]

    Augment the SGD optimizer to take structured inputs

    >>> new_sgd = structured_sgd(sgd)

    Data

    >>> y = np.linspace(1, 10, 100) + np.random.randn(100) + 1
    >>> X = np.array([np.ones(100), np.linspace(1, 100, 100)]).T
    >>> data = np.hstack((y[:, np.newaxis], X))

    Constant Initial values

    >>> w_0 = Parameter(np.array([1., 1.]), Bound())
    >>> lambda_0 = Parameter(.25, Bound())

    >>> res = new_sgd(cost, [w_0, lambda_0], data, batch_size=10,
    ...               eval_obj=True)
    >>> res_w, res_lambda = res.x

    Random Initial values

    >>> from scipy.stats import norm, gamma
    >>> w_0 = Parameter(norm(), Bound(), shape=(2,))
    >>> lambda_0 = Parameter(gamma(1.), Bound())

    >>> res = new_sgd(cost, [w_0, lambda_0], data, batch_size=10,
    ...               eval_obj=True, nstarts=100)
    >>> res_w, res_lambda = res.x
    """
    @wraps(sgd)
    def new_sgd(fun,
                parameters,
                data,
                eval_obj=False,
                batch_size=10,
                args=(),
                random_state=None,
                nstarts=100,
                **sgd_kwargs):

        (array1d, fbounds), shapes = flatten(parameters,
                                             hstack=bt.hstack,
                                             shape=bt.shape,
                                             ravel=bt.ravel
                                             )

        flatten_args_dec = flatten_args(shapes)
        new_fun = flatten_args_dec(fun)

        # Find best random starting candidate if we are doing random starts
        if eval_obj and nstarts > 0:
            data_gen = gen_batch(data, batch_size, random_state=random_state)
            array1d = _random_starts(
                fun=fun,
                parameters=parameters,
                jac=True,
                args=args,
                data_gen=data_gen,
                nstarts=nstarts,
                random_state=random_state
            )

        if bool(eval_obj):
            new_fun = flatten_func_grad(new_fun)
        else:
            new_fun = flatten_grad(new_fun)

        result = sgd(new_fun, array1d, data=data, bounds=fbounds, args=args,
                     eval_obj=eval_obj, random_state=random_state,
                     **sgd_kwargs)

        result['x'] = tuple(unflatten(result['x'], shapes))
        return result

    return new_sgd


def logtrick_minimizer(minimizer):
    r"""
    Log-Trick decorator for optimizers.

    This decorator implements the "log trick" for optimizing positive bounded
    variables. It will apply this trick for any variables that correspond to a
    Positive() bound.

    Examples
    --------
    >>> from scipy.optimize import minimize as sp_min
    >>> from ..btypes import Bound, Positive

    Here is an example where we may want to enforce a particular parameter or
    parameters to be strictly greater than zero,

    >>> def cost(w, lambda_):
    ...     sq_norm = w.T.dot(w)
    ...     return .5 * lambda_ * sq_norm, lambda_ * w

    Now let's enforce that the `w` are positive,

    >>> bounds = [Positive(), Positive(), Positive()]
    >>> new_min = logtrick_minimizer(sp_min)

    Initial values

    >>> w_0 = np.array([.5, .1, .2])
    >>> lambda_0 = .25

    >>> res = new_min(cost, w_0, args=(lambda_0,), bounds=bounds,
    ...               method='L-BFGS-B', jac=True)
    >>> res.x >= 0
    array([ True,  True,  True], dtype=bool)

    Note
    ----
    This decorator only works on unstructured optimizers. However, it can be
    use with structured_minimizer, so long as it is the inner wrapper.
    """
    @wraps(minimizer)
    def new_minimizer(fun, x0, jac=True, bounds=None, **minimizer_kwargs):

        if bounds is None:
            return minimizer(fun, x0, jac=jac, bounds=bounds,
                             **minimizer_kwargs)

        logx, expx, gradx, bounds = _logtrick_gen(bounds)

        # Intercept gradient
        if callable(jac):
            def new_jac(x, *fargs, **fkwargs):
                return gradx(jac(expx(x), *fargs, **fkwargs), x)
        else:
            new_jac = jac

        # Intercept objective
        if (not callable(jac)) and bool(jac):
            def new_fun(x, *fargs, **fkwargs):
                o, g = fun(expx(x), *fargs, **fkwargs)
                return o, gradx(g, x)
        else:
            def new_fun(x, *fargs, **fkwargs):
                return fun(expx(x), *fargs, **fkwargs)

        # Transform the final result
        result = minimizer(new_fun, logx(x0), jac=new_jac, bounds=bounds,
                           **minimizer_kwargs)
        result['x'] = expx(result['x'])
        return result

    return new_minimizer


def logtrick_sgd(sgd):
    r"""
    Log-Trick decorator for stochastic gradients.

    This decorator implements the "log trick" for optimizing positive bounded
    variables using SGD. It will apply this trick for any variables that
    correspond to a Positive() bound.

    Examples
    --------
    >>> from ..optimize import sgd
    >>> from ..btypes import Bound, Positive

    Here is an example where we may want to enforce a particular parameter or
    parameters to be strictly greater than zero,

    >>> def cost(w, data, lambda_):
    ...     N = len(data)
    ...     y, X = data[:, 0], data[:, 1:]
    ...     y_est = X.dot(w)
    ...     ww = w.T.dot(w)
    ...     obj = (y - y_est).sum() / N + lambda_ * ww
    ...     gradw = - 2 * X.T.dot(y - y_est) / N + 2 * lambda_ * w
    ...     return obj, gradw

    Now let's enforce that the `w` are positive,

    >>> bounds = [Positive(), Positive()]
    >>> new_sgd = logtrick_sgd(sgd)

    Data

    >>> y = np.linspace(1, 10, 100) + np.random.randn(100) + 1
    >>> X = np.array([np.ones(100), np.linspace(1, 100, 100)]).T
    >>> data = np.hstack((y[:, np.newaxis], X))

    Initial values

    >>> w_0 = np.array([1., 1.])
    >>> lambda_0 = .25

    >>> res = new_sgd(cost, w_0, data, args=(lambda_0,), bounds=bounds,
    ...               batch_size=10, eval_obj=True)
    >>> res.x >= 0
    array([ True,  True], dtype=bool)

    Note
    ----
    This decorator only works on unstructured optimizers. However, it can be
    use with structured_minimizer, so long as it is the inner wrapper.
    """
    @wraps(sgd)
    def new_sgd(fun, x0, data, bounds=None, eval_obj=False, **sgd_kwargs):

        if bounds is None:
            return sgd(fun, x0, data, bounds=bounds, eval_obj=eval_obj,
                       **sgd_kwargs)

        logx, expx, gradx, bounds = _logtrick_gen(bounds)

        if bool(eval_obj):
            def new_fun(x, *fargs, **fkwargs):
                o, g = fun(expx(x), *fargs, **fkwargs)
                return o, gradx(g, x)
        else:
            def new_fun(x, *fargs, **fkwargs):
                return gradx(fun(expx(x), *fargs, **fkwargs), x)

        # Transform the final result
        result = sgd(new_fun, logx(x0), data, bounds=bounds, eval_obj=eval_obj,
                     **sgd_kwargs)
        result['x'] = expx(result['x'])
        return result

    return new_sgd


#
# Helper functions
#


def flatten_grad(func):
    r"""
    Decorator to flatten structured gradients.

    Examples
    --------
    >>> def cost(w, lambda_):
    ...     sq_norm = w.T.dot(w)
    ...     return lambda_ * w, .5 * sq_norm
    >>> grad = cost(np.array([.5, .1, -.2]), .25)

    >>> len(grad)
    2
    >>> grad_w, grad_lambda = grad
    >>> np.shape(grad_w)
    (3,)
    >>> np.shape(grad_lambda)
    ()
    >>> grad_w
    array([ 0.125,  0.025, -0.05 ])
    >>> np.isclose(grad_lambda, 0.15)
    True

    >>> cost_new = flatten_grad(cost)
    >>> grad_new = cost_new(np.array([.5, .1, -.2]), .25)
    >>> grad_new
    array([ 0.125,  0.025, -0.05 ,  0.15 ])
    """
    @wraps(func)
    def new_func(*args, **kwargs):
        return flatten(func(*args, **kwargs), returns_shapes=False)

    return new_func


def flatten_func_grad(func):
    r"""
    Decorator to flatten structured gradients and return objective.

    Examples
    --------
    >>> def cost(w, lambda_):
    ...     sq_norm = w.T.dot(w)
    ...     return .5 * lambda_ * sq_norm, [lambda_ * w, .5 * sq_norm]
    >>> val, grad = cost(np.array([.5, .1, -.2]), .25)

    >>> np.isclose(val, 0.0375)
    True

    >>> len(grad)
    2
    >>> grad_w, grad_lambda = grad
    >>> np.shape(grad_w)
    (3,)
    >>> np.shape(grad_lambda)
    ()
    >>> grad_w
    array([ 0.125,  0.025, -0.05 ])
    >>> np.isclose(grad_lambda, 0.15)
    True

    >>> cost_new = flatten_func_grad(cost)
    >>> val_new, grad_new = cost_new(np.array([.5, .1, -.2]), .25)
    >>> val == val_new
    True
    >>> grad_new
    array([ 0.125,  0.025, -0.05 ,  0.15 ])
    """
    @wraps(func)
    def new_func(*args, **kwargs):
        val, grad = func(*args, **kwargs)
        return val, flatten(grad, returns_shapes=False)

    return new_func


def flatten_args(shapes):
    r"""
    Decorator to flatten structured arguments to a function.

    Examples
    --------
    >>> @flatten_args([(5,), ()])
    ... def f(w, lambda_):
    ...     return .5 * lambda_ * w.T.dot(w)
    >>> np.isclose(f(np.array([2., .5, .6, -.2, .9, .2])), .546)
    True
    >>> w = np.array([2., .5, .6, -.2, .9])
    >>> lambda_ = .2
    >>> np.isclose(.5 * lambda_ * w.T.dot(w), .546)
    True

    Some other curious applications

    >>> from operator import mul
    >>> flatten_args_dec = flatten_args([(), (3,)])
    >>> func = flatten_args_dec(mul)
    >>> func(np.array([3.1, .6, 1.71, -1.2]))
    array([ 1.86 ,  5.301, -3.72 ])
    >>> 3.1 * np.array([.6, 1.71, -1.2])
    array([ 1.86 ,  5.301, -3.72 ])
    >>> flatten_args_dec = flatten_args([(9,), (15,)])
    >>> func = flatten_args_dec(np.meshgrid)
    >>> x, y = func(np.arange(-5, 7, .5)) # 7 - (-5) / 0.5 = 24 = 9 + 15
    >>> x.shape
    (15, 9)
    >>> x[0, :]
    array([-5. , -4.5, -4. , -3.5, -3. , -2.5, -2. , -1.5, -1. ])
    >>> y.shape
    (15, 9)
    >>> y[:, 0]
    array([-0.5,  0. ,  0.5,  1. ,  1.5,  2. ,  2.5,  3. ,  3.5,  4. ,  4.5,
            5. ,  5.5,  6. ,  6.5])
    """
    def flatten_args_dec(func):

        @wraps(func)
        def new_func(array1d, *args, **kwargs):
            args = tuple(unflatten(array1d, shapes)) + args
            return func(*args, **kwargs)

        return new_func

    return flatten_args_dec


#
# Private module functions
#

def _random_starts(fun,
                   parameters,
                   jac,
                   args,
                   nstarts,
                   random_state,
                   data_gen=None):
    """Generate and evaluate random starts for Parameter objects."""
    if nstarts < 1:
        raise ValueError("nstarts has to be greater than or equal to 1")

    # Check to see if there are any random parameter types
    anyrand = any(flatten(map_recursive(lambda p: p.is_random, parameters),
                  returns_shapes=False))

    if not anyrand:
        log.info("No random parameters, not doing any random starts")
        params = map_recursive(lambda p: p.value, parameters, output_type=list)
        return params

    log.info("Evaluating random starts...")

    # Deal with gradient returns from objective function
    if jac is True:
        call_fun = lambda *fargs: fun(*fargs)[0]
    else:  # No gradient returns or jac is callable
        call_fun = fun

    # Randomly draw parameters and evaluate function
    def fun_eval():
        batch = next(data_gen) if data_gen else ()
        params = map_recursive(lambda p: p.rvs(random_state), parameters,
                               output_type=list)
        obj = call_fun(*chain(params, batch, args))
        return obj, params

    # Test randomly drawn parameters
    sample_gen = (fun_eval() for _ in range(nstarts))
    obj, params = min(sample_gen, key=lambda t: t[0])

    log.info("Best start found with objective = {}".format(obj))

    return flatten(params, returns_shapes=False)


def _logtrick_gen(bounds):
    """Generate warping functions and new bounds for the log trick."""
    # Test which parameters we can apply the log trick too
    ispos = np.array([isinstance(b, bt.Positive) for b in bounds], dtype=bool)
    nispos = ~ispos

    # Functions that implement the log trick
    def logx(x):
        xwarp = np.empty_like(x)
        xwarp[ispos] = np.log(x[ispos])
        xwarp[nispos] = x[nispos]
        return xwarp

    def expx(xwarp):
        x = np.empty_like(xwarp)
        x[ispos] = np.exp(xwarp[ispos])
        x[nispos] = xwarp[nispos]
        return x

    def gradx(grad, xwarp):
        gwarp = np.empty_like(grad)
        gwarp[ispos] = grad[ispos] * np.exp(xwarp[ispos])
        gwarp[nispos] = grad[nispos]
        return gwarp

    # Redefine bounds as appropriate for new ranges for numerical stability
    for i, (b, pos) in enumerate(zip(bounds, ispos)):
        if pos:
            upper = EXPMAX if b.upper is None else np.log(b.upper)
            bounds[i] = bt.Bound(lower=LOGMINPOS, upper=upper)

    return logx, expx, gradx, bounds
